"""Run all tests"""

import unittest
import sys


def run_tests():
    """Run all tests in current directory with signature test*.py"""
    testsuite = unittest.TestLoader().discover('', pattern="test*.py")
    run = unittest.TextTestRunner(verbosity=2).run(testsuite)
    return run.wasSuccessful()


if __name__ == '__main__':
    successful = run_tests()
    sys.exit(not successful)



"Subclass of Model for deployed models"
import requests
import numpy as np
from functools import partial

from .base import ModelType


class DeployedModel(ModelType):
    """Model that gets predictions from a deployed model"""
    def __init__(self, uri, input_formatter, output_formatter,
                 target_names=None, unique_values=None,
                 examples=None, feature_names=None,
                 request_kwargs={}, log_level=30):
        """This model can be called by making http requests
        to the passed in uri.

        Parameters
        ----------
        uri: string
            Where to post requests

        input_formatter: callable
            This function will run on input data before passing
            to requests library. This usually should take array types
            and convert them to JSON.

        output_formatter: callable
            This function will run on outputs before returning
            results to interpretation objects. This usually should take
            request objects and convert them to array types.

        target_names: array type
            see skater.model.Model for details

        unique_values: array type
            The set of possible output values. Only use on classifier models that
            return "best guess" predictions, not probability scores, e.g.

            model.predict(fruit1) -> 'apple'
            model.predict(fruit2) -> 'banana'

            ['apple','banana'] are the unique_values of the classifier

        examples:
            optional examples to use to make inferences about the function.

        request_kwargs: dict
            any additional request headers that need to be passed, such as api
            keys, content types, etc.

        log_level: int
            see skater.model.Model for details
        """
        self.uri = uri
        self.input_formatter = input_formatter
        self.output_formatter = output_formatter
        self.request_kwargs = request_kwargs
        super(DeployedModel, self).__init__(examples=examples,
                                            target_names=target_names,
                                            input_formatter=input_formatter,
                                            output_formatter=output_formatter,
                                            log_level=log_level,
                                            unique_values=unique_values,
                                            feature_names=feature_names)


    def _execute(self, data):
        """
        Just use the function itself for predictions
        """
        return requests.post(self.uri, json=data, **self.request_kwargs)


    @staticmethod
    def default_input_wrapper(data, key='input'):
        return {key: data.tolist()}


    @staticmethod
    def default_output_wrapper(response, key='prediction'):
        return np.array(response.json()[key])


    @staticmethod
    def _predict(data, uri, input_formatter, output_formatter, request_kwargs={}, transformer=None):
        """Static prediction function for multiprocessing usecases

        Parameters
        ----------
        data: arraytype

        uri: string
            Where to post requests

        input_formatter: callable
            This function will run on input data before passing
            to requests library. This usually should take array types
            and convert them to JSON.

        output_formatter: callable
            This function will run on outputs before returning
            results to interpretation objects. This usually should take
            request objects and convert them to array types.

        formatter: callable
            function responsible for formatting model outputs as necessary. For instance,
            one hot encoding multiclass outputs.

        predict_fn: callable

        Returns
        -----------
        predictions: arraytype
        """
        query = input_formatter(data)
        response = requests.post(uri, json=query, **request_kwargs)
        results = output_formatter(response)
        if transformer:
            results = transformer(results)
        return results


    def _get_static_predictor(self):

        uri = self.uri
        input_formatter = self.input_formatter
        output_formatter = self.output_formatter
        transformer = self.transformer
        predict_fn = partial(self._predict,
                             uri=uri,
                             input_formatter=input_formatter,
                             output_formatter=output_formatter,
                             transformer=transformer)
        return predict_fn



"""DataSet object"""
from __future__ import division
import numpy as np
import pandas as pd
from sklearn.metrics.pairwise import cosine_distances

from ..util.logger import build_logger
from ..util import exceptions
from ..util.static_types import StaticTypes
from ..util.dataops import add_column_numpy_array, allocate_samples_to_bins, flatten

__all__ = ['DataManager']


class DataManager(object):
    """Module for passing around data to interpretation objects"""

    # Todo: we can probably remove some of the keys from data_info, and have properties
    # executed as pure functions for easy to access metadata, such as n_rows, etc

    _n_rows = 'n_rows'
    _dim = 'dim'
    _feature_info = 'feature_info'
    _dtypes = 'dtypes'

    __attribute_keys__ = [_n_rows, _dim, _feature_info, _dtypes]
    __datatypes__ = (pd.DataFrame, np.ndarray)


    def __init__(self, data, feature_names=None, index=None, log_level=30):
        """
        The abstraction around using, accessing, sampling data for interpretation purposes.
        Used by interpretation objects to grab data, collect samples, and handle
        feature names and row indices.

        Parameters
        ----------
            data: 1D/2D numpy array, or pandas DataFrame
                raw data
            feature_names: iterable of feature names
                Optional keyword containing names of features.
            index: iterable of row names
                Optional keyword containing names of indexes (rows).

        """

        # create logger
        self._log_level = log_level
        self.logger = build_logger(log_level, __name__)

        if not isinstance(data, self.__datatypes__):
            err_msg = 'Invalid Data: expected data to be a numpy array or pandas dataframe but got ' \
                      '{}'.format(type(data))
            raise(exceptions.DataSetError(err_msg))

        ndim = len(data.shape)
        self.logger.debug("__init__ data.shape: {}".format(data.shape))

        if ndim == 1:
            data = data[:, np.newaxis]

        elif ndim >= 3:
            err_msg = "Invalid Data: expected data to be 1 or 2 dimensions, " \
                      "Data.shape: {}".format(ndim)
            raise(exceptions.DataSetError(err_msg))

        self.data = data
        self.data_type = type(data)
        self.metastore = None

        self.logger.debug("after transform data.shape: {}".format(self.data.shape))

        if isinstance(self.data, pd.DataFrame):
            if feature_names is None:
                feature_names = list(self.data.columns.values)
            if not index:
                index = list(self.data.index.values)
            self.feature_ids = list(feature_names)
            self.index = index

        elif isinstance(self.data, np.ndarray):
            if feature_names is None:
                feature_names = range(self.data.shape[1])
            if not index:
                index = range(self.data.shape[0])
            self.feature_ids = list(feature_names)
            self.index = index

        else:
            raise(ValueError("Invalid: currently we only support pandas dataframes and numpy arrays"
                             "If you would like support for additional data structures let us "
                             "know!"))

        self.data_info = {attr: None for attr in self.__attribute_keys__}


    def generate_grid(self, feature_ids, grid_resolution=100, grid_range=(.05, .95)):
        """
        Generates a grid of values on which to compute pdp. For each feature xi, for value
        yj of xi, we will fix xi = yj for every observation in X.

        Parameters
        ----------
            feature_ids(list):
                Feature names for which we'll generate a grid. Must be contained
                by self.feature_ids

            grid_resolution(int):
                The number of unique values to choose for each feature.

            grid_range(tuple):
                The percentile bounds of the grid. For instance, (.05, .95) corresponds to
                the 5th and 95th percentiles, respectively.

        Returns
        ----------
        grid(numpy.ndarray): 	There are as many rows as there are feature_ids
                                There are as many columns as specified by grid_resolution
        """

        if not all(i >= 0 and i <= 1 for i in grid_range):
            err_msg = "Grid range values must be between 0 and 1 but got:" \
                      "{}".format(grid_range)
            raise(exceptions.MalformedGridRangeError(err_msg))

        if not isinstance(grid_resolution, int) and grid_resolution > 0:
            err_msg = "Grid resolution {} is not a positive integer".format(grid_resolution)
            raise(exceptions.MalformedGridRangeError(err_msg))

        if not all(feature_id in self.feature_ids for feature_id in feature_ids):
            missing_features = []
            for feature_id in feature_ids:
                if feature_id not in self.feature_ids:
                    missing_features.append(feature_id)
            err_msg = "Feature ids {} not found in DataManager.feature_ids".format(missing_features)
            raise(KeyError(err_msg))

        grid_range = [x * 100 for x in grid_range]
        bins = np.linspace(*grid_range, num=grid_resolution).tolist()
        grid = []
        for feature_id in feature_ids:
            data = self[feature_id]
            info = self.feature_info[feature_id]
            # if a feature is categorical (non numeric) or
            # has a small number of unique values, we'll just
            # supply unique values for the grid
            if info['unique'] < grid_resolution or info['numeric'] is False:
                vals = np.unique(data)
            else:
                vals = np.unique(np.percentile(data, bins))
            grid.append(vals)
        grid = np.array(grid)
        grid_shape = [(1, i) for i in [row.shape[0] for row in grid]]
        self.logger.info('Generated grid of shape {}'.format(grid_shape))
        return grid


    def sync_metadata(self):
        self.data_info[self._n_rows] = self._calculate_n_rows()
        self.data_info[self._dim] = self._calculate_n_rows()
        self.data_info[self._dtypes] = self._calculate_dtypes()
        self.data_info[self._feature_info] = self._calculate_feature_info()


    def _calculate_n_rows(self):
        return self.data.shape[0]


    def _calculate_dim(self):
        return self.data.shape[1]


    def _calculate_dtypes(self):
        return pd.DataFrame(self.data, columns=self.feature_ids, index=self.index).dtypes


    def _calculate_feature_info(self):
        feature_info = {}
        for feature in self.feature_ids:
            x = self[feature]
            samples = self.generate_column_sample(feature, n_samples=10)
            samples_are_numeric = map(StaticTypes.data_types.is_numeric, samples)
            is_numeric = all(samples_are_numeric)
            feature_info[feature] = {
                'type': self.dtypes.loc[feature],
                'unique': len(np.unique(x)),
                'numeric': is_numeric
            }
        return feature_info


    @property
    def n_rows(self):
        if self.data_info[self._n_rows] is None:
            self.data_info[self._n_rows] = self._calculate_n_rows()
        return self.data_info[self._n_rows]


    @property
    def dim(self):
        if self.data_info[self._dim] is None:
            self.data_info[self._dim] = self._calculate_dim()
        return self.data_info[self._dim]


    @property
    def dtypes(self):
        if self.data_info[self._dtypes] is None:
            self.data_info[self._dtypes] = self._calculate_dtypes()
        return self.data_info[self._dtypes]


    @property
    def feature_info(self):
        if self.data_info[self._feature_info] is None:
            self.data_info[self._feature_info] = self._calculate_feature_info()
        return self.data_info[self._feature_info]


    def _build_metastore(self, bin_count):

        medians = np.median(np.array(self.data), axis=0).reshape(1, self.dim)

        # how far each data point is from the global median
        dists = cosine_distances(np.array(self.data), Y=medians).reshape(-1)

        # the percentile distance of each datapoint to the global median
        # dist_percentiles = map(lambda i: int(stats.percentileofscore(dists, i)), dists)

        bins = np.linspace(0, 100, num=bin_count + 1)
        unique_dists = np.unique(dists)

        if len(unique_dists) > 1:
            ranks_rounded = pd.qcut(dists, bins / 100, labels=False)
            unique_ranks = np.unique(ranks_rounded)
        else:
            ranks_rounded = np.ones(self.n_rows)
            unique_ranks = np.ones(1)
        return {
            'median': medians,
            'dists': dists,
            'n_rows': self.n_rows,
            'unique_ranks': unique_ranks,
            'ranks_rounded': ranks_rounded
        }


    def __getitem__(self, key):
        if issubclass(self.data_type, pd.DataFrame):
            return self.__getitem_pandas__(key)
        elif issubclass(self.data_type, np.ndarray):
            return self.__getitem_ndarray__(key)
        else:
            raise ValueError("Can't get item for data of type {}".format(self.data_type))


    def __setitem__(self, key, newval):
        if issubclass(self.data_type, pd.DataFrame):
            self.__setcolumn_pandas__(key, newval)
        elif issubclass(self.data_type, np.ndarray):
            self.__setcolumn_ndarray__(key, newval)
        else:
            raise ValueError("Can't set item for data of type {}".format(self.data_type))
        self.sync_metadata()


    def __getrows__(self, idx):
        if self.data_type == pd.DataFrame:
            return self.__getrows_pandas__(idx)
        elif self.data_type == np.ndarray:
            return self.__getrows_ndarray__(idx)
        else:
            raise ValueError("Can't get rows for data of type {}".format(self.data_type))


    def __getitem_pandas__(self, i):
        """if you passed in a pandas dataframe, it has columns which are strings."""
        return self.data[i]


    def __getitem_ndarray__(self, i):
        """if you passed in a pandas dataframe, it has columns which are strings."""
        if StaticTypes.data_types.is_string(i) or StaticTypes.data_types.is_numeric(i):
            idx = self.feature_ids.index(i)
        elif hasattr(i, '__iter__'):
            idx = [self.feature_ids.index(j) for j in i]
        else:
            raise(ValueError("Unrecongized index type: {}. This should not happen, "
                             "submit a issue here: "
                             "https://github.com/datascienceinc/Skater/issues"
                             .format(type(i))))
        return self.data[:, idx]


    def __setcolumn_pandas__(self, i, newval):
        """if you passed in a pandas dataframe, it has columns which are strings."""
        self.data[i] = newval


    def __setcolumn_ndarray__(self, i, newval):
        """if you passed in a pandas dataframe, it has columns which are strings."""
        if i in self.feature_ids:
            idx = self.feature_ids.index(i)
            self.data[:, idx] = newval
        else:
            self.data = add_column_numpy_array(self.data, newval)
            self.feature_ids.append(i)


    def __getrows_pandas__(self, idx):
        """if you passed in a pandas dataframe, it has columns which are strings."""
        return self.data.loc[idx]


    def __getrows_ndarray__(self, idx):
        """if you passed in a pandas dataframe, it has columns which are strings."""
        idx = [self.index.index(i) for i in idx]
        return self.data[idx]


    def generate_sample(self, sample=True, strategy='random-choice', n_samples_from_dataset=1000,
                        replace=True, samples_per_bin=10, bin_count=50):
        """
        Method for generating data from the dataset.

        Parameters:
        -----------
            sample(Bool):
                If False, we'll take the full dataset, otherwise we'll sample.

            n_samples_from_dataset(int):
                Specifies the number of samples to return. Only implemented
                if strategy is "random-choice".

            replace(Bool):
                Bool for sampling with or without replacement

            samples_per_bin(int):
                If strategy is uniform-over-similarity-ranks, then this is the number
                of samples to take from each discrete rank.


        """

        arg_dict = {
            'sample': sample,
            'strategy': strategy,
            'n_samples_from_dataset': n_samples_from_dataset,
            'replace': replace,
            'samples_per_bin': samples_per_bin,
            'bin_count': bin_count
        }
        self.logger.debug("Generating sample with args:\n {}".format(arg_dict))

        if not sample:
            return self.data

        if strategy == 'random-choice':
            idx = np.random.choice(self.index, size=n_samples_from_dataset, replace=replace)
            values = self.__getrows__(idx)
            return values

        elif strategy == 'uniform-from-percentile':
            raise(NotImplementedError("We havent coded this yet."))

        elif strategy == 'uniform-over-similarity-ranks':
            metastore = self._build_metastore(bin_count)
            data_distance_ranks = metastore['ranks_rounded']
            unique_ranks = metastore['unique_ranks']

            samples = []

            for rank in unique_ranks:
                idx = np.where(data_distance_ranks == rank)[0]
                if idx.any():
                    new_samples_idx = np.random.choice(idx, replace=True, size=samples_per_bin)
                    new_samples = self.__getrows__(new_samples_idx)
                    samples.extend(new_samples)
            if self.data_type == pd.DataFrame:
                return pd.DataFrame(samples, columns=self.feature_ids)
            elif self.data_type == np.ndarray:
                return np.array(samples)
            else:
                self.logger.warn("Type {} not in explictely supported. "
                                 "Returning sample as list".format(self.data_type))
                return samples

    def generate_column_sample(self, feature_id, n_samples=None, method='random-choice'):
        """Sample a single feature from the data set.

        Parameters
        ----------

        feature_id: hashable
            name of the feature to sample. If no feature names were passed, then
            the features are accessible via their column index.

        n_samples: int
            the number of samples to generate

        method: str
            the sampling method. Currently only random-choice is implemented.


        """
        if method == 'random-choice':
            return self._generate_column_sample_random_choice(feature_id, n_samples=n_samples)
        elif method == 'stratified':
            return self._generate_column_sample_stratified(feature_id, n_samples=n_samples)
        else:
            raise(NotImplementedError("Currenly we only support random-choice, stratified for "
                                      "column level sampling, not {} ".format(method)))

    def _generate_column_sample_random_choice(self, feature_id, n_samples=None):
        return np.random.choice(self[feature_id], size=n_samples)

    def _generate_column_sample_stratified(self, feature_id, n_samples=None, n_bins=100):
        """
        Tries to capture all relevant regions of space, relative to how many samples are allowed.
        Parameters:
        ----------
        feature_id:
        n_samples:

        Returns:
        ---------
        samples
        """
        if not self.data_info['feature_info'][feature_id]['numeric']:
            raise exceptions.DataSetError("Stratified sampling is currently "
                                          "supported for numeric features only.")

        bin_count, samples_per_bin = allocate_samples_to_bins(n_samples, ideal_bin_count=n_bins)
        percentiles = [100 * (i / bin_count) for i in range(bin_count + 1)]

        bins = list(np.percentile(self[feature_id], percentiles))
        sample_windows = [(bins[i], bins[i + 1]) for i in range(len(bins) - 1)]

        samples = []
        for window, n in zip(sample_windows, samples_per_bin):
            samples.append(np.random.uniform(window[0], window[1], size=int(n)).tolist())

        return np.array(flatten(samples))

    def _generate_column_sample_modeled(self, feature_id, n_samples=None):
        pass



"""Static types class and methods for inferring types."""
import numpy as np
import six


class ModelTypes(object):
    """Stores values for model types and keywords"""
    regressor = 'regressor'
    classifier = 'classifier'
    unknown = 'unknown'


class OutputTypes(object):
    """Stores values for output types, and keywords"""
    float = 'float'
    int = 'int'
    string = 'string'
    iterable = 'iterable'
    numeric = 'numeric'
    unknown = 'unknown'


class DataTypes(object):

    @staticmethod
    def is_numeric(thing):
        try:
            float(thing)
            return True
        except ValueError:
            return False
        except TypeError:
            return False


    @staticmethod
    def is_string(thing):
        return isinstance(thing, (six.text_type, six.binary_type))


    @staticmethod
    def is_dtype_numeric(dtype):
        assert isinstance(dtype, np.dtype), "expect numpy.dtype, got {}".format(type(dtype))
        return np.issubdtype(dtype, np.number)


class StaticTypes(object):
    """Stores values for model types, output types, and keywords"""
    model_types = ModelTypes
    output_types = OutputTypes
    data_types = DataTypes
    unknown = 'unknown'
    not_applicable = 'not applicable'


def return_data_type(thing):
    """Returns an output type given a variable"""
    if isinstance(thing, (six.text_type, six.binary_type)):
        return StaticTypes.output_types.string
    elif isinstance(thing, int):
        return StaticTypes.output_types.int
    elif isinstance(thing, float):
        return StaticTypes.output_types.float
    elif DataTypes.is_numeric(thing):
        return StaticTypes.output_types.numeric
    elif hasattr(thing, "__iter__"):
        return StaticTypes.output_types.iterable
    else:
        return StaticTypes.unknown



def pickle_method(method):
    func_name = method.im_func.__name__
    obj = method.im_self
    cls = method.im_class
    # deal with mangled names
    if func_name.startswith('__') and not func_name.endswith('__'):
        cls_name = cls.__name__.lstrip('_')
        func_name = '_' + cls_name + func_name
    return unpickle_method, (func_name, obj, cls)


def unpickle_method(func_name, obj, cls):
    for cls in cls.__mro__:
        try:
            func = cls.__dict__[func_name]
        except KeyError:
            pass
        else:
            break
    return func.__get__(obj, cls)



"""Utility for parsing command line arguments for testing"""

import argparse


def arg_parse(args):
    parser = create_parser()
    return parser.parse_args(args)


def create_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument('--debug', action='store_true')
    parser.add_argument('--seed', default=1, type=int)
    parser.add_argument('--n', default=1000, type=int)
    parser.add_argument('--dim', default=3, type=int)
    parser.add_argument('--r_deploy_test', action='store_true')
    parser.add_argument('--python_deploy_test', action='store_true')
    return parser



# Reference: http://python-packaging.readthedocs.io/en/latest/dependencies.htmlfrom setuptools import setup, find_packages
# meant to define package metadata in one and only one place

__title__ = 'skater'
__version__ = '1.0.0b.11'
__summary__ = 'Model Agnostic Interpretation Library'
__uri__ = 'https://github.com/datascienceinc/skater/'
__author__ = 'Aaron Kramer, DataScience.com Team, Pramit Choudhary'
__email__ = 'dstm@datascience.com'
__license__ = 'MIT'
__docs__ = 'https://datascienceinc.github.io/Skater/'



import unittest

import numpy as np
from scipy.stats import norm
from scipy.special import expit
from sklearn.linear_model import LinearRegression, LogisticRegression
from sklearn.ensemble import GradientBoostingClassifier
from sklearn import datasets
from functools import partial

from skater.core.explanations import Interpretation
from skater.util import exceptions
from arg_parser import create_parser
from skater.model import InMemoryModel, DeployedModel


class TestFeatureImportance(unittest.TestCase):

    def setUp(self):
        args = create_parser().parse_args()
        debug = args.debug
        self.seed = args.seed
        self.n = args.n
        self.dim = args.dim
        self.features = [str(i) for i in range(self.dim)]
        self.X = norm.rvs(0, 1, size=(self.n, self.dim), random_state=self.seed)
        self.B = np.array([-10.1, 2.2, 6.1])
        self.y = np.dot(self.X, self.B)
        self.y_as_int = np.round(expit(self.y))
        self.y_as_string = np.array([str(i) for i in self.y_as_int])
        # example dataset for y = B.X
        # X = array([[ 1.62434536, -0.61175641, -0.52817175], ... [-0.15065961, -1.40002289, -1.30106608]])  (1000 * 3)
        # B = array([-10.1,   2.2,   6.1])
        # y = array([ -2.09736000e+01,  -1.29850618e+00,  -1.73511155e+01, ...]) (1000 * 1)
        # features = ['0', '1', '2']
        ##
        # Other output types:
        # y_as_int = array[ 0.,  0.,  0.,  0.,  1.,  1.,  0.,  0.,  0.,  1.,  1.,  1.,  1., ...]
        # y_as_string = array['0.0', '0.0', '0.0', '0.0', '1.0', '1.0', '0.0', '0.0', '0.0', ... ]


        # Another set of input
        # sample data
        self.sample_x = np.array([[-2, -1], [-1, -1], [-1, -2], [1, 1], [1, 2], [2, 1]])
        self.sample_y = np.array([-1, -1, -1, 1, 1, 1])
        self.sample_feature_name = [str(i) for i in range(self.sample_x.shape[1])]

        if debug:
            self.interpreter = Interpretation(log_level='DEBUG')
        else:
            self.interpreter = Interpretation()  # default level is 'WARNING'
        self.interpreter.load_data(self.X, feature_names=self.features)

        self.regressor = LinearRegression()
        self.regressor.fit(self.X, self.y)
        self.regressor_predict_fn = InMemoryModel(self.regressor.predict, examples=self.X)

        self.classifier = LogisticRegression()
        self.classifier.fit(self.X, self.y_as_int)
        self.classifier_predict_fn = InMemoryModel(self.classifier.predict, examples=self.X, unique_values=self.classifier.classes_)
        self.classifier_predict_proba_fn = InMemoryModel(self.classifier.predict_proba, examples=self.X)

        self.string_classifier = LogisticRegression()
        self.string_classifier.fit(self.X, self.y_as_string)
        self.string_classifier_predict_fn = InMemoryModel(self.string_classifier.predict_proba, examples=self.X)



    @staticmethod
    def feature_column_name_formatter(columnname):
        return "feature: {}".format(columnname)


    def test_feature_importance(self):
        importances = self.interpreter.feature_importance.feature_importance(self.regressor_predict_fn)
        self.assertEquals(np.isclose(importances.sum(), 1), True)  # default grid resolution is 100


    def test_plot_feature_importance(self):
        self.interpreter.feature_importance.plot_feature_importance(self.regressor_predict_fn)


if __name__ == '__main__':
    suite = unittest.TestLoader().loadTestsFromTestCase(TestFeatureImportance)
    unittest.TextTestRunner(verbosity=2).run(suite)



import unittest

import numpy as np
import pandas as pd
from numpy.testing import assert_array_equal
from scipy.special import expit
from sklearn.linear_model import LinearRegression, LogisticRegression
from functools import partial

from skater.core.local_interpretation.lime.lime_tabular import LimeTabularExplainer


class TestLime(unittest.TestCase):
    """
    Test imported lime package
    """

    def setUp(self):
        """
        Build data for testing
        :param n:
        :param dim:
        :return:
        """

        self.X = np.array([
            [0, 0, 0],
            [0, 0, 1],
            [0, 1, 0],
            [0, 1, 1],
            [1, 0, 0],
            [1, 0, 1],
            [1, 1, 0],
            [1, 1, 1],
        ])
        self.n, self.dim = self.X.shape
        self.feature_names = ['x{}'.format(i) for i in range(self.dim)]
        self.index = ['{}'.format(i) for i in range(self.n)]

        self.B = np.array([-5, 0, 5])
        self.y = np.dot(self.X, self.B) + np.random.normal(0, .01, size=self.n)
        self.y_for_classifier = np.round(expit(self.y))
        self.example = self.X[0]

        self.seed = 1
        self.regressor = LinearRegression()
        self.regressor.fit(self.X, self.y)

        self.classifier = LogisticRegression()
        self.classifier.fit(self.X, self.y_for_classifier)

        self.model_regressor = LinearRegression()

    def test_regression_with_feature_names(self):
        """
        Ensure lime.lime_tabular works when predict_fn = regressor.predict
        and feature names are passed
        :return:
        """

        interpretor = LimeTabularExplainer(self.X, feature_names=self.feature_names)
        assert interpretor.explain_regressor_instance(self.example, self.regressor.predict)

    def test_regression_without_feature_names(self):
        """
        Ensure lime.lime_tabular works when predict_fn = regressor.predict
        and feature names are NOT passed
        :return:
        """
        interpretor = LimeTabularExplainer(self.X)
        assert interpretor.explain_regressor_instance(self.example, self.regressor.predict)

    def test_classifier_no_proba_without_feature_names(self):
        """
        Ensure lime.lime_tabular works when predict_fn = classifier.predict
        and feature names are NOT passed
        :return:
        """

        interpretor = LimeTabularExplainer(self.X)
        interpretor_func = partial(interpretor.explain_instance, *[self.example, self.classifier.predict])
        self.assertRaises(NotImplementedError, interpretor_func)

    def test_classifier_with_proba_without_feature_names(self):
        """
        Ensure lime.lime_tabular works when predict_fn = classifier.predict_proba
        and feature names are NOT passed
        :return:
        """

        interpretor = LimeTabularExplainer(self.X)
        assert interpretor.explain_instance(self.example, self.classifier.predict_proba)

    def test_classifier_no_proba_with_feature_names(self):
        """
        Ensure lime.lime_tabular works when predict_fn = classifier.predict
        and feature names are passed
        :return:
        """

        interpretor = LimeTabularExplainer(self.X, feature_names=self.feature_names)
        interpretor_func = partial(interpretor.explain_instance, *[self.example, self.classifier.predict])
        self.assertRaises(NotImplementedError, interpretor_func)

    def test_classifier_with_proba_with_feature_names(self):
        """
        Ensure lime.lime_tabular works when predict_fn = classifier.predict_proba
        and feature names are passed
        :return:
        """

        interpretor = LimeTabularExplainer(self.X, feature_names=self.feature_names)
        assert interpretor.explain_instance(self.example, self.classifier.predict_proba)

    def test_lime_coef_accuracy(self):
        """
        Ensure that for a trivial example, the coefficients of a regressor explanation
        are all similar to the true beta values of the generative process.

        :return:
        """

        error_epsilon = .1
        explainer = LimeTabularExplainer(self.X,
                                         discretize_continuous=True)
        explanation = explainer.explain_regressor_instance(self.example,
                                                           self.regressor.predict,
                                                           model_regressor=self.model_regressor)

        vals = dict(explanation.as_list())
        keys = ['{} <= 0.00'.format(i) for i in [2, 1, 0]]
        lime_coefs = np.array([vals[key] for key in keys])
        assert (abs(self.regressor.coef_ - lime_coefs) < error_epsilon).all()



if __name__ == '__main__':
    runner = unittest.TextTestRunner(verbosity=2)
    runner.run(unittest.makeSuite(TestLime))



import unittest

import numpy as np
import pandas as pd
from numpy.testing import assert_array_equal

from skater.data import DataManager
from arg_parser import arg_parse, create_parser


class TestData(unittest.TestCase):
    """
    Tests the skater.data.DataManager object
    """

    def setUp(self):
        """Create data for testing"""
        self.parser = create_parser()
        args = self.parser.parse_args()
        debug = args.debug

        self.dim = 3
        self.n = 8

        self.X = np.array([
            [0, 0, 0],
            [0, 0, 2],
            [0, 1, 0],
            [0, 2, 1],
            [1, 0, 0],
            [1, 0, 1],
            [1, 1, 0],
            [2, 1, 1],
        ])

        self.feature_names = ['x{}'.format(i) for i in range(self.dim)]
        self.index = ['{}'.format(i) for i in range(self.n)]

        if debug:
            self.log_level = 10
        else:
            self.log_level = 30


    def test_datamanager_data_returns_original(self):
        """
        ensure DataManager(data).data == data
        """
        data_set = DataManager(self.X,
                               feature_names=self.feature_names,
                               index=self.index)
        assert_array_equal(data_set.data, self.X)


    def test_1d_numpy_array(self):
        """
        ensure DataManager(data) works when data is 1D np.ndarray
        """

        feature_id = 0
        array_1d = self.X[:, feature_id][:, np.newaxis]
        feature_names = [self.feature_names[feature_id]]

        data_set = DataManager(array_1d,
                               feature_names=feature_names,
                               index=self.index,
                               log_level=self.log_level)
        assert_array_equal(data_set.data, array_1d)
        assert data_set.feature_ids == feature_names


    def test_2d_numpy_array(self):
        """
        ensure DataManager(data) works when data is 2D np.ndarray
        """

        feature_ids = [0, 1]
        array_2d = self.X[:, feature_ids]
        feature_names = [self.feature_names[i] for i in feature_ids]

        data_set = DataManager(array_2d,
                               feature_names=feature_names,
                               index=self.index,
                               log_level=self.log_level)
        assert_array_equal(data_set.data, array_2d)
        assert data_set.feature_ids == feature_names


    def test_pandas_dataframe(self):
        """
        Ensure DataManager(data) works when data is pd.DataFrame
        """
        X_as_dataframe = pd.DataFrame(self.X, columns=self.feature_names, index=self.index)
        data_set = DataManager(X_as_dataframe, log_level=self.log_level)

        assert data_set.feature_ids == self.feature_names, "Feature Names from DataFrame " \
                                                           "not loaded properly"
        assert data_set.index == self.index, "Index from DataFrame not loaded properly"
        assert_array_equal(data_set.data, self.X)


    def test_generate_grid_1_variable(self):
        """Ensures generate grid works with 1 variable"""
        data_set = DataManager(self.X, feature_names=self.feature_names, index=self.index)
        grid = data_set.generate_grid(data_set.feature_ids[0:1], grid_resolution=100)
        self.assertEquals(len(grid), 1)


    def test_generate_grid_2_variables(self):
        """Ensures generate grid works with 2 variables"""
        data_set = DataManager(self.X, feature_names=self.feature_names, index=self.index)
        grid = data_set.generate_grid(self.feature_names[0:2], grid_resolution=100)
        self.assertEquals(len(grid), 2)


if __name__ == '__main__':
    runner = unittest.TextTestRunner(verbosity=2)
    runner.run(unittest.makeSuite(TestData))



"""Partial Dependence class"""

from itertools import product, cycle
import numpy as np
import pandas as pd
from pathos.multiprocessing import Pool
import functools

from ...data import DataManager
from .base import BaseGlobalInterpretation
from ...model.base import ModelType
from ...util import exceptions
from ...util.user_defined_types import ControlledDict
from ...util.kernels import flatten
from ...util.plotting import COLORS, \
    coordinate_gradients_to_1d_colorscale, plot_2d_color_scale
from ...util.exceptions import *
from ...util.static_types import StaticTypes

# if we want to employ instance methods in multiprocessing, enable this code:
# copy_reg.pickle(types.MethodType, pickle_method, unpickle_method)
# methods stored in util.serialization


def _compute_pd(index, estimator_fn, grid_expanded, pd_metadata, input_data, filter_classes=None):
    """ Helper function to compute partial dependence for each grid value. This function is
    designed to unbound/static to avoid issues when computing partial dendendences in
    separate processes.

    Parameters:
    -----------
    index(int): row index for the grid
    estimator_fn(estimator.function):
        an estimator function of a fitted model used to derive prediction.
        Supports classification(binary, multi-class) and regression.
    grid_expanded(numpy.ndarray:
        The grid of ``target_labels` for which partial dependence needs to be computed
    number_of_classes(int):
        unique number of classes in the ``target_labels``
    feature_ids(list):
        the names/ids of the features for which partial dependence is to be computed.
    input_data(numpy.ndarray):
        input sample data as array to compute partial dependence

    Returns
    -------
    pd_dict(dict, shape={'sd': <>, 'val_1': <>, 'mean'} : containing estimated value on sample dataset
    """

    # column names for the target features in the partial dependence results
    feature_columns = pd_metadata['feature_columns_for_pd']

    # feature_ids for the target features in the partial dependence results
    feature_ids = pd_metadata['feature_ids_for_pd']

    # column names of the target classes in the partial dependence results
    target_columns = pd_metadata['target_names']

    # all of the feature_ids in the data set
    all_feature_ids = list(pd_metadata['all_feature_ids'])

    # values of the target features to set
    new_row = grid_expanded[index]
    number_of_classes = len(target_columns)

    # create a copy so mutations dont have side effects, and using a
    # datamanager for consistent __setitem__ calls.
    data_set = DataManager(input_data.copy(), feature_names=all_feature_ids)
    for feature_idx, feature_id in enumerate(feature_ids):
        data_set[feature_id] = new_row[feature_idx]

    # generate statistics for the new predictions
    predictions = estimator_fn(data_set.data)
    mean_prediction = np.mean(predictions, axis=0)
    std_prediction = np.std(predictions, axis=0)

    # Todo: add static version of model.predict_subset_classes, use here
    if len(predictions.shape) == 1:
        mean_prediction = np.array([mean_prediction])
        std_prediction = np.array([std_prediction])

    if filter_classes is not None:
        class_idx = [target_columns.index(i) for i in filter_classes]
        mean_prediction = mean_prediction[class_idx]
        target_columns = [target_columns[i] for i in class_idx]
        number_of_classes = len(filter_classes)

    pd_dict = {column: new_row[idx] for idx, column in enumerate(feature_columns)}

    # if binary classification and user has not explicitly asked for 2 classes
    # just return results for one class
    if number_of_classes == 2 and filter_classes is None:
        target_column = target_columns[1]
        pd_dict[target_column] = mean_prediction[1]
        pd_dict['sd'] = std_prediction[0]
    else:
        for class_i in range(number_of_classes):
            pd_dict[target_columns[class_i]] = mean_prediction[class_i]
        pd_dict['sd'] = std_prediction[0]

    return pd_dict


class PartialDependence(BaseGlobalInterpretation):
    """Contains methods for partial dependence. Subclass of BaseGlobalInterpretation

       Partial dependence adapted from:

       T. Hastie, R. Tibshirani and J. Friedman,
       Elements of Statistical Learning Ed. 2, Springer, 2009.
    """

    __all__ = ['partial_dependence', 'plot_partial_dependence']

    def _build_metadata_dict(self, modelinstance, pd_feature_ids, data_feature_ids, filter_classes):

        feature_columns = [self.feature_column_name_formatter(i) for i in pd_feature_ids]
        sd_col = 'sd'
        if filter_classes is not None:
            filtered_target_names = [i for i in modelinstance.target_names if i in filter_classes]
        else:
            filtered_target_names = None
        metadata = ControlledDict({
            'sd_column': sd_col,
            'target_names': modelinstance.target_names,
            'filtered_target_names': filtered_target_names,
            'feature_columns_for_pd': feature_columns,
            'feature_ids_for_pd': pd_feature_ids,
            'all_feature_ids': data_feature_ids,
        })
        metadata.block_setitem()
        return metadata

    @staticmethod
    def feature_column_name_formatter(columnname):
        return "{}".format(columnname)

    def _check_features(self, feature_ids):
        if not hasattr(feature_ids, "__iter__"):
            feature_ids = [feature_ids]

        if len(feature_ids) >= 3:
            too_many_features_err_msg = "Pass in at most 2 features for pdp. If you have a " \
                                        "use case where you'd like to look at 3 simultaneously" \
                                        ", please let us know."
            raise(exceptions.TooManyFeaturesError(too_many_features_err_msg))

        if len(feature_ids) == 0:
            empty_features_err_msg = "Feature ids must have non-zero length"
            raise(exceptions.EmptyFeatureListError(empty_features_err_msg))

        if len(set(feature_ids)) != len(feature_ids):
            duplicate_features_error_msg = "feature_ids cannot contain duplicate values"
            raise(exceptions.DuplicateFeaturesError(duplicate_features_error_msg))

        return feature_ids

    def partial_dependence(self, feature_ids, modelinstance, filter_classes=None, grid=None,
                           grid_resolution=30, n_jobs=-1, grid_range=None, sample=True,
                           sampling_strategy='random-choice', n_samples=1000,
                           bin_count=50, samples_per_bin=10, return_metadata=False):

        """
        Approximates the partial dependence of the predict_fn with respect to the
        variables passed.

        Parameters:
        -----------
        feature_ids: list
            the names/ids of the features for which partial dependence is to be computed.
            Note that the algorithm's complexity scales exponentially with additional
            features, so generally one should only look at one or two features at a
            time. These feature ids must be available in the class's associated DataSet.
            As of now, we only support looking at 1 or 2 features at a time.
        modelinstance: skater.model.model.Model subtype
            an estimator function of a fitted model used to derive prediction. Supports
            classification and regression. Supports classification(binary, multi-class) and regression.
            predictions = predict_fn(data)

            Can either by a skater.model.remote.DeployedModel or a
            skater.model.local.InMemoryModel
        filter_classes: array type
            The classes to run partial dependence on. Default None invokes all classes.
            Only used in classification models.
        grid: numpy.ndarray
            2 dimensional array on which we fix values of features. Note this is
            determined automatically if not given based on the percentiles of the
            dataset.
        grid_resolution: int
            how many unique values to include in the grid. If the percentile range
            is 5% to 95%, then that range will be cut into <grid_resolution>
            equally size bins. Defaults to 30.
        n_jobs: int
            The number of CPUs to use to compute the PDs. -1 means 'all CPUs'.
            Defaults to using all cores(-1).
        grid_range: tuple
            the percentile extrama to consider. 2 element tuple, increasing, bounded
            between 0 and 1.
        sample: boolean
            Whether to sample from the original dataset.
        sampling_strategy: string
            If sampling, which approach to take. See DataSet.generate_sample for
            details.
        n_samples: int
            The number of samples to use from the original dataset. Note this is
            only active if sample = True and sampling strategy = 'uniform'. If
            using 'uniform-over-similarity-ranks', use samples per bin
        bin_count: int
            The number of bins to use when using the similarity based sampler. Note
            this is only active if sample = True and
            sampling_strategy = 'uniform-over-similarity-ranks'.
            total samples = bin_count * samples per bin.
        samples_per_bin: int
            The number of samples to collect for each bin within the sampler. Note
            this is only active if sample = True and
            sampling_strategy = 'uniform-over-similarity-ranks'. If using
            sampling_strategy = 'uniform', use n_samples.
            total samples = bin_count * samples per bin.
        return_metadata: boolean

        :Example:
        >>> from skater.model import InMemoryModel
        >>> from skater.core.explanations import Interpretation
        >>> from sklearn.ensemble import RandomForestClassier
        >>> from sklearn.datasets import load_boston
        >>> boston = load_boston()
        >>> X = boston.data
        >>> y = boston.target
        >>> features = boston.feature_names

        >>> rf = RandomForestClassier()
        >>> rf.fit(X,y)


        >>> model = InMemoryModel(rf, examples = X)
        >>> interpreter = Interpretation()
        >>> interpreter.load_data(X)
        >>> feature_ids = ['ZN','CRIM']
        >>> interpreter.partial_dependence.partial_dependence(features,model)
        """

        if self.data_set is None:
            load_data_not_called_err_msg = "self.interpreter.data_set not found. " \
                                           "Please call Interpretation.load_data " \
                                           "before running this method."
            raise(exceptions.DataSetNotLoadedError(load_data_not_called_err_msg))

        feature_ids = self._check_features(feature_ids)

        if filter_classes:
            err_msg = "members of filter classes must be" \
                      "members of modelinstance.classes." \
                      "Expected members of: " \
                      "{0}\n" \
                      "got: " \
                      "{1}".format(modelinstance.target_names,
                                   filter_classes)
            assert all([i in modelinstance.target_names for i in filter_classes]), err_msg

        # TODO: There might be a better place to do this check
        if not isinstance(modelinstance, ModelType):
            raise(exceptions.ModelError("Incorrect estimator function used for computing partial dependence, try one "
                                        "creating one with skater.model.local.InMemoryModel or"
                                        "skater.model.remote.DeployedModel"))

        if modelinstance.model_type == 'classifier' and modelinstance.probability is False:

            if modelinstance.unique_values is None:
                raise(exceptions.ModelError('If using classifier without probability scores, unique_values cannot '
                                            'be None'))
            self.interpreter.logger.warn("Classifiers with probability scores can be explained "
                                         "more granularly than those without scores. If a prediction method with "
                                         "scores is available, use that instead.")

        # TODO: This we can change easily to functional style
        missing_feature_ids = []
        for feature_id in feature_ids:
            if feature_id not in self.data_set.feature_ids:
                missing_feature_ids.append(feature_id)

        if missing_feature_ids:
            missing_feature_id_err_msg = "Features {0} not found in " \
                                         "Interpretation.data_set.feature_ids" \
                                         "{1}".format(missing_feature_ids, self.data_set.feature_ids)
            raise(KeyError(missing_feature_id_err_msg))

        if grid_range is None:
            grid_range = (.05, 0.95)
        else:
            if not hasattr(grid_range, "__iter__"):
                err_msg = "Grid range {} needs to be an iterable".format(grid_range)
                raise(exceptions.MalformedGridRangeError(err_msg))

        self._check_grid_range(grid_range)

        if not modelinstance.has_metadata:
            examples = self.data_set.generate_sample(strategy='random-choice',
                                                     sample=True,
                                                     n_samples_from_dataset=10)
            examples = DataManager(examples, feature_names=self.data_set.feature_ids)
            modelinstance._build_model_metadata(examples)

        # if you dont pass a grid, build one.
        grid = np.array(grid)
        if not grid.any():
            # Currently, if a given feature has fewer unique values than the value
            # of grid resolution, then the grid will be set to those unique values.
            # Otherwise it will take the percentile
            # range according with grid_resolution bins.
            grid = self.data_set.generate_grid(feature_ids,
                                               grid_resolution=grid_resolution,
                                               grid_range=grid_range)
        else:
            # want to ensure all grids have 2 axes
            if len(grid.shape) == 1 and \
                    (StaticTypes.data_types.is_string(grid[0]) or StaticTypes.data_types.is_numeric(grid[0])):
                grid = grid[:, np.newaxis].T
                grid_resolution = grid.shape[1]

        self.interpreter.logger.debug("Grid shape used for pdp: {}".format(grid.shape))
        self.interpreter.logger.debug("Grid resolution for pdp: {}".format(grid_resolution))

        # make sure data_set module is giving us correct data structure
        self._check_grid(grid, feature_ids)

        # generate data
        data_sample = self.data_set.generate_sample(strategy=sampling_strategy,
                                                    sample=sample,
                                                    n_samples_from_dataset=n_samples,
                                                    samples_per_bin=samples_per_bin,
                                                    bin_count=bin_count)

        _pdp_metadata = self._build_metadata_dict(modelinstance, feature_ids, self.data_set.feature_ids, filter_classes)

        self.interpreter.logger.debug("Shape of sampled data: {}".format(data_sample.shape))
        self.interpreter.logger.debug("Feature Ids: {}".format(feature_ids))
        self.interpreter.logger.debug("PD metadata: {}".format(_pdp_metadata))

        # cartesian product of grid
        grid_expanded = pd.DataFrame(list(product(*grid))).values

        if grid_expanded.shape[0] <= 0:
            empty_grid_expanded_err_msg = "Must have at least 1 pdp value" \
                                          "grid shape: {}".format(grid_expanded.shape)
            raise(exceptions.MalformedGridError(empty_grid_expanded_err_msg))

        predict_fn = modelinstance._get_static_predictor()

        n_jobs = None if n_jobs < 0 else n_jobs
        pd_func = functools.partial(_compute_pd,
                                    estimator_fn=predict_fn,
                                    grid_expanded=grid_expanded,
                                    pd_metadata=_pdp_metadata,
                                    input_data=data_sample,
                                    filter_classes=filter_classes)
        arg_list = [i for i in range(grid_expanded.shape[0])]
        executor_instance = Pool(n_jobs)
        try:
            pd_list = executor_instance.map(pd_func, arg_list)
        except:
            self.interpreter.logger.debug("Multiprocessing failed, going single process")
            pd_list = map(pd_func, arg_list)
        finally:
            executor_instance.close()
            executor_instance.join()
            executor_instance.terminate()
        if return_metadata:
            return pd.DataFrame(list(pd_list)), _pdp_metadata
        else:
            return pd.DataFrame(list(pd_list))


    def plot_partial_dependence(self, feature_ids, modelinstance, filter_classes=None,
                                grid=None, grid_resolution=30, grid_range=None,
                                n_jobs=-1, sample=True, sampling_strategy='random-choice',
                                n_samples=1000, bin_count=50, samples_per_bin=10,
                                with_variance=False, figsize=(16, 10)):
        """
        Computes partial_dependence of a set of variables. Essentially approximates
        the partial partial_dependence of the predict_fn with respect to the variables
        passed.

        Parameters:
        -----------
        feature_ids: list
            the names/ids of the features for which partial dependence is to be computed.
            Note that the algorithm's complexity scales exponentially with additional
            features, so generally one should only look at one or two features at a
            time. These feature ids must be available in the class's associated DataSet.
            As of now, we only support looking at 1 or 2 features at a time.
        modelinstance: skater.model.model.Model subtype
            an estimator function of a fitted model used to derive prediction. Supports
            classification and regression. Supports classification(binary, multi-class) and regression.
            predictions = predict_fn(data)

            Can either by a skater.model.remote.DeployedModel or a
            skater.model.local.InMemoryModel
        grid: numpy.ndarray
            2 dimensional array on which we fix values of features. Note this is
            determined automatically if not given based on the percentiles of the
            dataset.
        grid_resolution: int
            how many unique values to include in the grid. If the percentile range
            is 5% to 95%, then that range will be cut into <grid_resolution>
            equally size bins. Defaults to 30.
        grid_range: tuple
            the percentile extrama to consider. 2 element tuple, increasing, bounded
            between 0 and 1.
        n_jobs: int
            The number of CPUs to use to compute the PDs. -1 means 'all CPUs'.
            Defaults to using all cores(-1).
        sample: boolean
            Whether to sample from the original dataset.
        sampling_strategy: string
            If sampling, which approach to take. See DataSet.generate_sample for
            details.
        n_samples: int
            The number of samples to use from the original dataset. Note this is
            only active if sample = True and sampling strategy = 'uniform'. If
            using 'uniform-over-similarity-ranks', use samples per bin
        bin_count: int
            The number of bins to use when using the similarity based sampler. Note
            this is only active if sample = True and
            sampling_strategy = 'uniform-over-similarity-ranks'.
            total samples = bin_count * samples per bin.
        samples_per_bin: int
            The number of samples to collect for each bin within the sampler. Note
            this is only active if sample = True and
            sampling_strategy = 'uniform-over-similarity-ranks'. If using
            sampling_strategy = 'uniform', use n_samples.
            total samples = bin_count * samples per bin.


        with_variance(bool):
            whether to include pdp error bars in the plots. Currently disabled for 3D
            plots for visibility. If you have a use case where you'd like error bars for
            3D pdp plots, let us know!
        plot_title(string):
            title for pdp plots

        Example
        --------
        >>> from sklearn.ensemble import GradientBoostingRegressor
        >>> from sklearn.datasets.california_housing import fetch_california_housing
        >>> cal_housing = fetch_california_housing()
        # split 80/20 train-test
        >>> x_train, x_test, y_train, y_test = train_test_split(cal_housing.data,
        >>>                             cal_housing.target, test_size=0.2, random_state=1)
        >>> names = cal_housing.feature_names
        >>> print("Training the estimator...")
        >>> estimator = GradientBoostingRegressor(n_estimators=10, max_depth=4,
        >>>                             learning_rate=0.1, loss='huber', random_state=1)
        >>> estimator.fit(x_train, y_train)
        >>> from skater.core.explanations import Interpretation
        >>> interpreter = Interpretation()
        >>> print("Feature name: {}".format(names))
        >>> interpreter.load_data(X_train, feature_names=names)
        >>> print("Input feature name: {}".format[names[1], names[5]])
        >>> from skater.model import InMemoryModel
        >>> model = InMemoryModel(clf.predict, examples = X_train)
        >>> interpreter.partial_dependence.plot_partial_dependence([names[1], names[5]], model,
        >>>                                                         n_samples=100, n_jobs=1)

        """

        try:
            global pyplot
            global ScalarFormatter
            global Axes3D
            global mpl_axes
            global cm
            global tick_formatter
            from matplotlib.axes._subplots import Axes as mpl_axes
            # from matplotlib.ticker import ScalarFormatter
            from mpl_toolkits.mplot3d import Axes3D
            from matplotlib import pyplot, cm
            from ...util.plotting import tick_formatter
        except ImportError:
            raise (MatplotlibUnavailableError("Matplotlib is required but unavailable on your system."))
        except RuntimeError:
            raise (MatplotlibDisplayError("Matplotlib unable to open display"))

        # in the event that a user wants a 3D pdp with multiple classes, how should
        # we handle this? currently each class will get its own figure
        if not hasattr(feature_ids, "__iter__"):
            pd_df, metadata = self.partial_dependence(feature_ids, modelinstance,
                                                      filter_classes=filter_classes, grid=grid,
                                                      grid_resolution=grid_resolution,
                                                      grid_range=grid_range, sample=sample,
                                                      sampling_strategy=sampling_strategy,
                                                      n_samples=n_samples, bin_count=bin_count,
                                                      samples_per_bin=samples_per_bin,
                                                      n_jobs=n_jobs, return_metadata=True)

            self.interpreter.logger.info("done computing pd, now plotting ...")
            ax = self._plot_pdp_from_df(pd_df, metadata, with_variance=with_variance, figsize=figsize)
            return ax
        else:
            ax_list = []
            for feature_or_feature_pair in feature_ids:
                pd_df, metadata = self.partial_dependence(feature_or_feature_pair, modelinstance,
                                                          filter_classes=filter_classes, grid=grid,
                                                          grid_resolution=grid_resolution,
                                                          grid_range=grid_range, sample=sample,
                                                          sampling_strategy=sampling_strategy,
                                                          n_samples=n_samples, bin_count=bin_count,
                                                          samples_per_bin=samples_per_bin,
                                                          n_jobs=n_jobs, return_metadata=True)

                self.interpreter.logger.info("done computing pd, now plotting ...")
                ax = self._plot_pdp_from_df(pd_df, metadata, with_variance=with_variance, figsize=figsize)
                ax_list.append(ax)
            return ax_list


    def _plot_pdp_from_df(self, pdp, pd_metadata,
                          with_variance=False, plot_title=None,
                          disable_offset=True, figsize=(16, 10)):

        feature_columns = pd_metadata['feature_columns_for_pd']
        if pd_metadata['filtered_target_names'] is None:
            target_columns = pd_metadata['target_names']
        else:
            target_columns = pd_metadata['filtered_target_names']
        sd_col = pd_metadata['sd_column']
        n_features = len(feature_columns)
        if n_features == 1 or not hasattr(feature_columns, "__iter__"):
            feature_column = feature_columns[0]
            return self._2d_pdp_plot(pdp,
                                     feature_column,
                                     sd_col,
                                     target_columns,
                                     with_variance=with_variance,
                                     plot_title=plot_title,
                                     disable_offset=disable_offset,
                                     figsize=figsize)
        elif n_features == 2:
            feature1_column, feature2_column = feature_columns
            return self._3d_pdp_plot(pdp,
                                     feature1_column,
                                     feature2_column,
                                     sd_col,
                                     target_columns,
                                     with_variance=with_variance,
                                     plot_title=plot_title,
                                     figsize=figsize)
        else:
            msg = "Something went wrong. Expected either a single feature, " \
                  "or a 1-2 element array of features, got array of size:" \
                  "{}: {}".format(*[n_features, feature_columns])
            raise(ValueError(msg))


    def _2d_pdp_plot(self, pdp, feature_name, sd_col, target_columns,
                     with_variance=False, plot_title=None,
                     disable_offset=True, figsize=(16, 10)):
        colors = cycle(COLORS)
        figure_list, axis_list = [], []

        # if there are just 2 classes, pick the last one.
        if len(target_columns) == 2:
            target_columns = [target_columns[-1]]

        for target_column in target_columns:
            # if target_name is None:
            #     raise ValueError("Could not parse class name from {}".format(mean_col))
            f, ax = pyplot.subplots(1, figsize=figsize)
            figure_list.append(f)
            axis_list.append(ax)
            color = next(colors)

            data = pdp.set_index(feature_name)
            plane = data[target_column]

            # if binary feature, then len(pdp) == 2 -> barchart
            if self._is_feature_binary(pdp, feature_name) or not self.data_set.feature_info[feature_name]['numeric']:
                if with_variance:
                    error = data[sd_col]
                else:
                    error = None
                plane.plot(kind='bar', ax=ax, color=color, yerr=error)
            else:
                plane.plot(ax=ax, color=color)
                if with_variance:
                    upper_plane = plane + data[sd_col]
                    lower_plane = plane - data[sd_col]
                    ax.fill_between(data.index.values,
                                    lower_plane.values,
                                    upper_plane.values,
                                    alpha=.2,
                                    color=color)
            if plot_title:
                ax.set_title(plot_title)
            ax.set_ylabel(target_column)
            ax.set_xlabel(feature_name)
            handles, labels = ax.get_legend_handles_labels()
            ax.legend(handles, labels)
            if disable_offset:
                ax.yaxis.set_major_formatter(tick_formatter())
        return flatten([figure_list, axis_list])


    def _is_feature_binary(self, pdp, feature):
        data = pdp[feature].values
        if len(np.unique(data)) == 2:
            return True
        else:
            return False


    def _3d_pdp_plot(self, pdp, feature1, feature2, sd_column, target_columns,
                     with_variance=False, plot_title=None, disable_offset=True,
                     figsize=(16, 10)):

        # if there are just 2 classes, pick the last one.
        if len(target_columns) == 2:
            target_columns = [target_columns[-1]]

        feature1_n_uniques = self.data_set.feature_info[feature1]['unique']
        feature2_n_uniques = self.data_set.feature_info[feature2]['unique']
        feature1_numeric = self.data_set.feature_info[feature1]['numeric']
        feature2_numeric = self.data_set.feature_info[feature2]['numeric']

        feature_1_is_categorical = (feature1_n_uniques == 2) or not feature1_numeric
        feature_2_is_categorical = (feature2_n_uniques == 2) or not feature2_numeric

        if not feature_1_is_categorical and not feature_2_is_categorical:
            self.interpreter.logger.debug("Neither feature is binary, so plotting 3D mesh")
            plot_objects = self._plot_3d_full_mesh(pdp,
                                                   feature1,
                                                   feature2,
                                                   sd_column,
                                                   target_columns,
                                                   with_variance=with_variance,
                                                   figsize=figsize)

        elif feature_1_is_categorical and feature_2_is_categorical:
            self.interpreter.logger.debug("Both features are binary, so plotting groups")
            plot_objects = self._plot_2d_2_categorical_features_bar(pdp,
                                                                    feature1,
                                                                    feature2,
                                                                    sd_column,
                                                                    target_columns,
                                                                    with_variance=with_variance,
                                                                    figsize=figsize)
        else:
            # one feature is binary and one isnt.
            categorical_feature, non_categorical_feature = {
                True: [feature1, feature2],
                False: [feature2, feature1]
            }[feature_1_is_categorical]
            self.interpreter.logger.debug("One feature is categorical, and one isnt")
            self.interpreter.logger.debug("Categorical Feature: {}".format(categorical_feature))
            self.interpreter.logger.debug("Non Categorical Feature: {}".format(non_categorical_feature))

            plot_objects = self._plot_2d_1_categorical_feature_and_1_continuous(pdp,
                                                                                categorical_feature,
                                                                                non_categorical_feature,
                                                                                sd_column,
                                                                                target_columns,
                                                                                with_variance=with_variance)
        for obj in plot_objects:
            if isinstance(obj, mpl_axes):
                if disable_offset:

                    xlabels = [i.get_text() for i in obj.get_xticklabels()]
                    if all(StaticTypes.data_types.is_numeric(i) for i in xlabels):
                        obj.xaxis.set_major_formatter(tick_formatter())
                    obj.yaxis.set_major_formatter(tick_formatter())
                if plot_title:
                    obj.set_title("Partial Dependence")
                # matplotlib increases x from left to right, flipping that
                # so the origin is front and center
        return plot_objects


    def _plot_3d_full_mesh(self, pdp, feature1, feature2,
                           sd_column, target_columns,
                           with_variance=False, alpha=.7, figsize=(16, 10)):
        colors = cycle(COLORS)

        figure_list, axis_list = [], []

        for target_column in target_columns:
            gradient_x, gradient_y, X, Y, Z = self.compute_3d_gradients(pdp, target_column, feature1, feature2)
            color_gradient, xmin, xmax, ymin, ymax = coordinate_gradients_to_1d_colorscale(gradient_x, gradient_y)
            figure = pyplot.figure(figsize=figsize)
            ax = pyplot.subplot2grid((3, 3), (0, 0), colspan=2, rowspan=3, projection='3d')
            figure_list.append(figure)
            axis_list.append(ax)

            ax.plot_surface(X, Y, Z, alpha=alpha, facecolors=color_gradient, linewidth=0., rstride=1, cstride=1)
            # in case we'd like to return these values to the user
            # dx_mean = np.mean(gradient_x)
            # dy_mean = np.mean(gradient_y)
            # mean_point = (dx_mean, dy_mean)

            # add 2D color scale
            ax_colors = pyplot.subplot2grid((3, 3), (1, 2), colspan=1, rowspan=1)
            ax_colors = plot_2d_color_scale(xmin, xmax, ymin, ymax, ax=ax_colors)
            ax_colors.set_xlabel("{}".format(feature1))
            ax_colors.set_ylabel("{}".format(feature2), rotation=270, labelpad=10)
            ax_colors.set_title("Gradient of PDP")
            ax_colors.yaxis.tick_right()
            ax_colors.yaxis.set_label_position("right")
            ax_colors.xaxis.set_major_formatter(tick_formatter())
            ax_colors.yaxis.set_major_formatter(tick_formatter())



            if with_variance:
                var_color = next(colors)
                ax.plot_trisurf(pdp[feature1].values, pdp[feature2].values,
                                (pdp[target_column] + pdp[sd_column]).values, alpha=.2,
                                color=var_color)
                ax.plot_trisurf(pdp[feature1].values, pdp[feature2].values,
                                (pdp[target_column] - pdp[sd_column]).values, alpha=.2,
                                color=var_color)
            ax.set_xlabel(feature1)
            ax.set_ylabel(feature2)
            # adding a blank line and spacing for formatting
            ax.set_zlabel("\n{}".format(target_column), linespacing=3.0)
            ax.invert_xaxis()

            handles, labels = ax.get_legend_handles_labels()
            ax.legend(handles, labels)

        return flatten([figure_list, axis_list])


    def _plot_3d_2_categorical_features(self,
                                        pdp,
                                        feature1,
                                        feature2,
                                        sd_column,
                                        target_columns,
                                        with_variance=False,
                                        figsize=(16, 10)):
        """This method is not currently implemented, and is incomplete. If enabled,
        we need to add handling for if with_variance.
        """
        figure_list, axis_list = [], []
        for target_column in target_columns:
            fig = pyplot.figure(figsize=figsize)
            ax = fig.add_subplot(111, projection='3d')

            for val in np.unique(pdp[feature2]):
                filter_idx = pdp[feature2] == val
                pdp_vals = pdp[filter_idx][target_column].values
                x1 = pdp[filter_idx][feature1].values
                x2 = pdp[filter_idx][feature2].values
                ax.plot(x1, x2, pdp_vals)

            figure_list.append(fig)
            axis_list.append(ax)
            # color = next(colors)
            ax.set_xlabel(feature1)
            ax.set_ylabel(feature2)
            ax.set_zlabel(target_column)
            handles, labels = ax.get_legend_handles_labels()
            ax.legend(handles, labels)
        return flatten([figure_list, axis_list])

    def _plot_2d_2_categorical_features_lines(self,
                                              pdp,
                                              feature1,
                                              feature2,
                                              sd_col,
                                              target_columns,
                                              with_variance=False,
                                              figsize=(16, 10)):
        figure_list, axis_list = [], []

        std_error = pdp.set_index([feature1, feature2])[sd_col].unstack()
        for target_column in target_columns:
            f = pyplot.figure(figsize=figsize)
            ax = f.add_subplot(111)
            # feature2 is columns
            # feature1 is index
            plot_data = pdp.set_index([feature1, feature2])[target_column].unstack()
            plot_data.plot(ax=ax, color=COLORS)

            if with_variance:
                colors = cycle(COLORS)
                binary1_values = plot_data.index.values
                binary2_values = plot_data.columns.values
                for binary2_value in binary2_values:
                    color = next(colors)
                    yerr = std_error[binary2_value].values
                    upper_plane = yerr + plot_data[binary2_value].values
                    lower_plane = plot_data[binary2_value].values - yerr
                    ax.fill_between(binary1_values, lower_plane, upper_plane,
                                    color=color, alpha=.2)
            figure_list.append(f)
            axis_list.append(ax)
            ax.set_xlabel(feature1)
            ax.set_ylabel(target_column)

        return flatten([figure_list, axis_list])

    def _plot_2d_2_categorical_features_bar(self,
                                            pdp,
                                            feature1,
                                            feature2,
                                            sd_col,
                                            target_columns,
                                            with_variance=False,
                                            figsize=(16, 10)):
        figure_list, axis_list = [], []

        std_error = pdp.set_index([feature1, feature2])[sd_col].unstack()
        for target_column in target_columns:
            f = pyplot.figure(figsize=figsize)
            ax = f.add_subplot(111)
            # feature2 is columns, feature1 is index
            plot_data = pdp.set_index([feature1, feature2])[target_column].unstack()

            if with_variance:
                plot_data.plot(kind='bar', ax=ax, color=COLORS, yerr=std_error)
            else:
                plot_data.plot(kind='bar', ax=ax, color=COLORS)

            ax.set_xticklabels(plot_data.index.values)
            figure_list.append(f)
            axis_list.append(ax)
            ax.set_xlabel(feature1)
            ax.set_ylabel(target_column)

        return flatten([figure_list, axis_list])


    def _plot_2d_1_categorical_feature_and_1_continuous(self,
                                                        pdp,
                                                        categorical_feature,
                                                        non_categorical_feature,
                                                        sd_column,
                                                        target_columns,
                                                        with_variance=False,
                                                        figsize=(16, 10)):

        figure_list, axis_list = [], []

        for target_column in target_columns:
            colors = cycle(COLORS)
            f = pyplot.figure(figsize=figsize)
            ax = f.add_subplot(111)
            figure_list.append(f)
            axis_list.append(ax)
            plot_data = pdp.set_index([non_categorical_feature, categorical_feature])[target_column]\
                .unstack().sort_index()
            sd = pdp.set_index([non_categorical_feature, categorical_feature])[sd_column]\
                .unstack()

            plot_data.plot(ax=ax, color=COLORS)
            if with_variance:
                non_categorical_values = map(float, plot_data.index.values)
                categorical_values = plot_data.columns.values
                upper_plane = plot_data + sd
                lower_plane = plot_data - sd
                for categorical_value in categorical_values:
                    color = next(colors)
                    ax.fill_between(non_categorical_values, lower_plane[categorical_value].values,
                                    upper_plane[categorical_value].values, alpha=.2, color=color)
            ax.set_ylabel(target_column)
        return flatten([figure_list, axis_list])

    @staticmethod
    def _check_grid(grid, feature_ids):
        if not isinstance(grid, np.ndarray):
            err_msg = "Grid of type {} must be a numpy array".format(type(grid))
            raise(exceptions.MalformedGridError(err_msg))

        if len(feature_ids) != grid.shape[0]:
            err_msg = "Given {0} features, there must be {1} rows in grid" \
                      "but {2} were found".format(len(feature_ids),
                                                  len(feature_ids),
                                                  grid.shape[0])
            raise(exceptions.MalformedGridError(err_msg))


    @staticmethod
    def _check_dataset(dataset):
        """
        Ensures that dataset is pandas dataframe, and dataset is not empty
        :param dataset: skater.__datatypes__
        :return:
        """
        if not isinstance(dataset, (pd.DataFrame, np.ndarray)):
            err_msg = "Dataset.data must be a pandas.DataFrame or numpy.ndarray"
            raise(exceptions.DataSetError(err_msg))

        if len(dataset) == 0:
            err_msg = "Dataset.data is empty"
            raise (exceptions.DataSetError(err_msg))


    @staticmethod
    def _check_grid_range(grid_range):
        """
        Tested by unit test, ensures grid range is between 0 and 1
        :param grid_range (tuple)

        """
        if len(grid_range) != 2:
            err_msg = "Grid range {} must have 2 elements".format(grid_range)
            raise(exceptions.MalformedGridRangeError(err_msg))
        if not all([i >= 0 and i <= 1 for i in grid_range]):
            err_msg = "All elements of grid range {} " \
                      "must be between 0 and 1".format(grid_range)
            raise(exceptions.MalformedGridRangeError(err_msg))

    @staticmethod
    def compute_3d_gradients(pdp, mean_col, feature_1, feature_2, scaled=True):
        """
        Computes component-wise gradients of pdp dataframe.

        Parameters
        ----------
        pdp: pandas.DataFrame
            DataFrame containing partial dependence values
        mean_col: string
            column name corresponding to pdp value
        feature_1: string
            column name corresponding to feature 1
        feature_2: string
            column name corresponding to feature 2
        scaled: bool
            Whether to scale the x1 and x2 gradients relative to x1 and x2 bin sizes

        Returns
        ----------
        dx, dy, x_matrix, y_matrix, z_matrix
        """
        def feature_vals_to_grad_deltas(values):
            values = np.unique(values)
            values.sort()
            diffs = np.diff(values)
            conv_diffs = np.array([(diffs[i] + diffs[i + 1]) / 2 for i in range(0, len(diffs) - 1)])
            diffs = np.concatenate((np.array([diffs[0]]), conv_diffs, np.array([diffs[-1]])))
            return diffs

        df = pdp.sort_values([feature_1, feature_2])

        feature_1_diffs = feature_vals_to_grad_deltas(df[feature_1].values)
        feature_2_diffs = feature_vals_to_grad_deltas(df[feature_2].values)

        x1_size = feature_1_diffs.shape[0]
        x2_size = feature_2_diffs.shape[0]

        z_matrix = np.zeros((x1_size, x2_size))
        x1_matrix = np.zeros((x1_size, x2_size))
        x2_matrix = np.zeros((x1_size, x2_size))

        for i in range(x1_size):
            for j in range(x2_size):
                idx = i * x2_size + j
                x1_val = df[feature_1].iloc[idx]
                x2_val = df[feature_2].iloc[idx]
                z_val = df[mean_col].iloc[idx]

                z_matrix[i, j] = z_val
                x1_matrix[i, j] = x1_val
                x2_matrix[i, j] = x2_val

        dx1, dx2 = np.gradient(z_matrix)
        if scaled:
            dx1 = np.apply_along_axis(lambda x: x / feature_1_diffs, 0, dx1)
            dx2 = np.apply_along_axis(lambda x: x / feature_2_diffs, 1, dx2)
        return dx1, dx2, x1_matrix, x2_matrix, z_matrix



"Specialized Exceptions for skater"


def exception_factory(exception_name, base_exception=Exception, attributes=None):
    attribute_dict = {
        "__init__": base_exception.__init__
    }
    if isinstance(attributes, dict):
        attributes.update(attributes)
    return type(
        exception_name,
        (base_exception, ),
        attribute_dict
    )


DataSetNotLoadedError = exception_factory('DataSetNotLoadedError')

PartialDependenceError = exception_factory('PartialDependenceError')

FeatureImportanceError = exception_factory('FeatureImportanceError')

DataSetError = exception_factory('DataSetError')

ModelError = exception_factory("ModelError")

TooManyFeaturesError = exception_factory('TooManyFeaturesError',
                                         base_exception=PartialDependenceError)

DuplicateFeaturesError = exception_factory('DuplicateFeaturesError',
                                           base_exception=PartialDependenceError)

EmptyFeatureListError = exception_factory('EmptyFeatureListError',
                                          base_exception=PartialDependenceError)

MalformedGridError = exception_factory("MalformedGridError",
                                       base_exception=PartialDependenceError)

MalformedGridRangeError = exception_factory("MalformedGridRangeError",
                                            base_exception=PartialDependenceError)

MatplotlibUnavailableError = exception_factory('MatplotlibUnavailableError', base_exception=ImportError)

MatplotlibDisplayError = exception_factory('MatplotlibDisplayError', base_exception=RuntimeError)



"""Feature Importance class"""
from itertools import cycle
import numpy as np
import pandas as pd

from ...data import DataManager
from .base import BaseGlobalInterpretation
from ...util.plotting import COLORS
from ...util.exceptions import *
from ...model.base import ModelType
from ...util.dataops import divide_zerosafe


class FeatureImportance(BaseGlobalInterpretation):
    """Contains methods for feature importance. Subclass of BaseGlobalInterpretation.

    """

    def feature_importance(self, model_instance, ascending=True, filter_classes=None):

        """
        Computes feature importance of all features related to a model instance.
        Supports classification, multi-class classification, and regression.

        Wei, Pengfei, Zhenzhou Lu, and Jingwen Song.
        "Variable Importance Analysis: A Comprehensive Review".
        Reliability Engineering & System Safety 142 (2015): 399-432.

        Parameters
        ----------
        model_instance: skater.model.model.Model subtype
            the machine learning model "prediction" function to explain, such that
            predictions = predict_fn(data).
        ascending: boolean, default True
            Helps with ordering Ascending vs Descending
        filter_classes: array type
            The classes to run partial dependence on. Default None invokes all classes.
            Only used in classification models.

        Returns
        -------
        importances : Sorted Series


        Examples
        --------
            >>> from skater.model import InMemoryModel
            >>> from skater.core.explanations import Interpretation
            >>> from sklearn.ensemble import RandomForestClassier
            >>> rf = RandomForestClassier()
            >>> rf.fit(X,y)
            >>> model = InMemoryModel(rf, examples = X)
            >>> interpreter = Interpretation()
            >>> interpreter.load_data(X)
            >>> interpreter.feature_importance.feature_importance(model)
        """

        if filter_classes:
            err_msg = "members of filter classes must be" \
                      "members of model_instance.classes." \
                      "Expected members of: {0}\n" \
                      "got: {1}".format(model_instance.target_names,
                                        filter_classes)
            assert all([i in model_instance.target_names for i in filter_classes]), err_msg

        importances = {}
        original_predictions = model_instance.predict_subset_classes(self.data_set.data, filter_classes)

        n = original_predictions.shape[0]

        copy_of_data_set = DataManager(self.data_set.data.copy(),
                                       feature_names=self.data_set.feature_ids,
                                       index=self.data_set.index)

        for feature_id in self.data_set.feature_ids:
            # collect perturbations
            if self.data_set.feature_info[feature_id]['numeric']:
                samples = self.data_set.generate_column_sample(feature_id, n_samples=n, method='stratified')
            else:
                samples = self.data_set.generate_column_sample(feature_id, n_samples=n, method='random-choice')
            copy_of_data_set[feature_id] = samples

            # predict based on perturbed values
            new_predictions = model_instance.predict_subset_classes(copy_of_data_set.data, filter_classes)

            importance = self.compute_importance(new_predictions,
                                                 original_predictions,
                                                 self.data_set[feature_id],
                                                 samples)
            importances[feature_id] = importance

            # reset copy
            copy_of_data_set[feature_id] = self.data_set[feature_id]

        importances = pd.Series(importances).sort_values(ascending=ascending)

        if not importances.sum() > 0:
            self.interpreter.logger.debug("Importances that caused a bug: {}".format(importances))
            raise(FeatureImportanceError("Something went wrong. Importances do not sum to a positive value"
                                         "This could be due to:"
                                         "1) 0 or infinite divisions"
                                         "2) perturbed values == original values"
                                         "3) feature is a constant"
                                         ""
                                         "Please submit an issue here:"
                                         "https://github.com/datascienceinc/Skater/issues"))

        importances = divide_zerosafe(importances, (np.ones(importances.shape[0]) * importances.sum()))
        return importances


    def plot_feature_importance(self, predict_fn, filter_classes=None, ascending=True, ax=None):
        """Computes feature importance of all features related to a model instance,
        then plots the results. Supports classification, multi-class classification, and regression.

        Parameters
        ----------
        predict_fn: skater.model.model.Model subtype
            estimator "prediction" function to explain the predictive model. Could be probability scores
            or target values
        filter_classes: array type
            The classes to run partial dependence on. Default None invokes all classes.
            Only used in classification models.
        ascending: boolean, default True
            Helps with ordering Ascending vs Descending
        ax: matplotlib.axes._subplots.AxesSubplot
            existing subplot on which to plot feature importance. If none is provided,
            one will be created.

        Returns
        -------
        f: figure instance
        ax: matplotlib.axes._subplots.AxesSubplot
            could be used to for further modification to the plots

        Examples
        --------
            >>> from skater.model import InMemoryModel
            >>> from skater.core.explanations import Interpretation
            >>> from sklearn.ensemble import RandomForestClassier
            >>> rf = RandomForestClassier()
            >>> rf.fit(X,y)
            >>> model = InMemoryModel(rf, examples = X)
            >>> interpreter = Interpretation()
            >>> interpreter.load_data(X)
            >>> interpreter.feature_importance.plot_feature_importance(model, ascending=True, ax=ax)
            """
        try:
            global pyplot
            from matplotlib import pyplot
        except ImportError:
            raise (MatplotlibUnavailableError("Matplotlib is required but unavailable on your system."))
        except RuntimeError:
            raise (MatplotlibDisplayError("Matplotlib unable to open display"))

        importances = self.feature_importance(predict_fn, filter_classes=filter_classes)

        if ax is None:
            f, ax = pyplot.subplots(1)
        else:
            f = ax.figure

        colors = cycle(COLORS)
        color = next(colors)
        # Below is a weirdness because of how pandas plot is behaving. There might be a better way
        # to resolve the issuse of sorting based on axis
        if ascending is True:
            importances.sort_values(ascending=False).plot(kind='barh', ax=ax, color=color)
        else:
            importances.sort_values(ascending=True).plot(kind='barh', ax=ax, color=color)
        return f, ax


    def compute_importance(self, new_predictions, original_predictions, original_x, perturbed_x,
                           method='output-variance', scaled=False):
        if method == 'output-variance':
            importance = self._compute_importance_via_output_variance(np.array(new_predictions),
                                                                      np.array(original_predictions),
                                                                      np.array(original_x),
                                                                      np.array(perturbed_x),
                                                                      scaled)
        else:
            raise(KeyError("Unrecongized method for computing feature_importance: {}".format(method)))
        return importance


    def _compute_importance_via_output_variance(self, new_predictions, original_predictions,
                                                original_x, perturbed_x, scaled=True):
        """Mean absolute error of predictions given perturbations in a feature"""
        changes_in_predictions = abs(new_predictions - original_predictions)

        if scaled:
            changes_in_predictions = self._importance_scaler(changes_in_predictions, original_x, perturbed_x)

        importance = np.mean(changes_in_predictions)
        return importance


    def _importance_scaler(self, values, original_x, perturbed_x):
        raise(NotImplementedError("We currently don't support scaling, we are researching the best"
                                  "approaches to do so."))



"""Interpretation Class"""

from .global_interpretation.partial_dependence import PartialDependence
from .global_interpretation.feature_importance import FeatureImportance
from ..data import DataManager
from ..util.logger import build_logger


class Interpretation(object):
    """
    Interpretation class. Before calling interpretation subclasses like partial
    dependence, one must call Interpretation.load_data().

    Attributes:
    ----------
    data_set: skater.data.DataSet
        skater's data abstraction layer. All interactions with data goes through
         this object.

    local_interpreter: skater.core.local_interpreter.LocalInterpreter
        Contains methods for interpreting single prediction. Currently exposes the lime
        library.

    global_interpreter: skater.core.global_interpreter.GlobalInterpreter
        Contains methods for evaluating a model over entire regions of the domain. Currently
        exposes partial dependency


    Examples:
    ----------

    from skater.core.explanations import Interpretation
    interpreter = Interpretation()
    interpreter.load_data(X, feature_ids = ['a','b'])
    interpreter.partial_dependence([feature_id1, feature_id2], regressor.predict)
    """

    def __init__(self, training_data=None, feature_names=None, index=None,
                 log_level=30):
        """
        Attaches local and global interpretations
        to Interpretation object.

        Parameters:
        -----------
        log_level: int
            Logger Verbosity, see https://docs.python.org/2/library/logging.html
            for details.

        """
        self._log_level = log_level
        self.logger = build_logger(log_level, __name__)
        self.partial_dependence = PartialDependence(self)
        self.feature_importance = FeatureImportance(self)
        self.data_set = None

        if training_data is not None:
            self.load_data(training_data, feature_names=feature_names, index=index)


    def load_data(self, training_data, feature_names=None, index=None):
        """
        Creates a DataSet object from inputs, ties to interpretation object.
        This will be exposed to all submodules.

        Parameters
        ----------
        training_data: numpy.ndarray, pandas.DataFrame
            the dataset. can be 1D or 2D

        feature_names: array-type
            names to call features.

        index: array-type
            names to call rows.


        Returns
        ---------
            None
        """

        self.logger.info("Loading Data")
        self.data_set = DataManager(training_data,
                                    feature_names=feature_names,
                                    index=index,
                                    log_level=self._log_level)
        self.logger.info("Data loaded")
        self.logger.debug("Data shape: {}".format(self.data_set.data.shape))
        self.logger.debug("Dataset Feature_ids: {}".format(self.data_set.feature_ids))



"""Base method for all global interpretations. Is
a subclass of base ModelInterpreter"""

from ..model_interpreter import ModelInterpreter


class BaseGlobalInterpretation(ModelInterpreter):
    """Base class for global model interpretations"""
    pass



"""Holds all the kernels"""
import numpy as np


def rbf_kernel(distance, kernel_width=1.0):
    """Passes a distance through a gaussian kernel"""
    return np.sqrt(np.exp(-(distance ** 2) / kernel_width ** 2))


def flatten(array):
    return [item for sublist in array for item in sublist]



"""Includes kernels, static type/properties for package, exceptions, logging."""
# from .kernels import rbf_kernel
# from .static_types import return_data_type, StaticTypes, ModelTypes, OutputTypes
# from .data_structures import ControlledDict
# from .model import get_predictor
# from .serialization import pickle_method, unpickle_method

__all__ = [
    'rbf_kernel',
    'StaticTypes',
    'return_data_type',
    'build_static_predictor',
    'ModelTypes',
    'OutputTypes',
    'ControlledDict',
    'pickle_method',
    'unpickle_method']



import numpy as np

from .exceptions import MatplotlibDisplayError, MatplotlibUnavailableError

COLORS = ['#328BD5', '#404B5A', '#3EB642', '#E04341', '#8665D0']


class LinearSegments(object):

    black_to_blue_dict = {
        'red': ((0.0, 0.0, 0.0),
                (1.0, 0.0, 0.0)),
        'blue': ((0.0, 0.0, 0.0),
                 (1.0, 1.0, 0.0)),
        'green': ((0.0, 0.0, 0.0),
                  (1.0, 0.0, 0.0))}

    black_to_green_dict = {
        'red': ((0.0, 0.0, 0.0),
                (1.0, 0.0, 0.0)),
        'blue': ((0.0, 0.0, 0.0),
                 (1.0, 0.0, 0.0)),
        'green': ((0.0, 0.0, 0.0),
                  (1.0, 1.0, 0.0))}

    red_to_green_dict = {
        'red': ((0.0, 1.0, 1.0),
                (1.0, 0.0, 0.0)),
        'blue': ((0.0, 0.0, 0.0),
                 (1.0, 0.0, 0.0)),
        'green': ((0.0, 0.0, 0.0),
                  (1.0, 1.0, 0.0))}


class ColorMap(object):
    """
    Maps arrays to colors
    """
    def __init__(self):
        try:
            global colors
            global cm
            from matplotlib import colors, cm
        except ImportError:
            raise (MatplotlibUnavailableError("Matplotlib is required but unavailable on your system."))
        except RuntimeError:
            raise (MatplotlibDisplayError("Matplotlib unable to open display"))

        self.red_to_green = colors.LinearSegmentedColormap('my_colormap', LinearSegments.red_to_green_dict, 100)
        self.black_to_blue = colors.LinearSegmentedColormap('my_colormap', LinearSegments.black_to_blue_dict, 100)
        self.black_to_green = colors.LinearSegmentedColormap('my_colormap', LinearSegments.black_to_green_dict, 100)


    def array_1d_to_color_scale(self, array_1d, colormap):
        mmin, mmax = min(array_1d), max(array_1d)
        norm = colors.Normalize(mmin, mmax)
        scalarMapx = cm.ScalarMappable(norm=norm, cmap=colormap)
        scalarMapx.set_array(array_1d)
        return scalarMapx.to_rgba(array_1d)


def coordinate_gradients_to_1d_colorscale(dx, dy,
                                          x_buffer_prop=.1, y_buffer_prop=.1,
                                          norm='separate'):
    """
    Map x and y gradients to single array of colors based on 2D color scale

    Parameters
    ----------
    dx: array type
        x component of gradient
    dy: array type
        y component of gradient
    x_buffer_prop:
    y_buffer_prop:
    norm: string
        whether to normalize colors based on differences in x and y scales.
        if separate, each component gets its own scaling (default)
        if global, will scale based on global extrema

    Returns
    ----------
    color_array, xmin, xmax, ymin, ymax
    """
    colormap = ColorMap()
    xmin, xmax, xbuffer = build_buffer(dx.min(), dx.max())
    ymin, ymax, ybuffer = build_buffer(dy.min(), dy.max())
    global_min = min(xmin, ymin)
    global_max = max(xmax, ymax)
    if norm == 'separate':
        normx = colors.Normalize(xmin, xmax)
        normy = colors.Normalize(ymin, ymax)
    elif norm == 'shared':
        normx = colors.Normalize(global_min, global_max)
        normy = colors.Normalize(global_min, global_max)
    else:
        raise KeyError("keyword norm must be in ('separate', 'shared')")

    scalarMapx = cm.ScalarMappable(norm=normx, cmap=colormap.black_to_blue)
    scalarMapy = cm.ScalarMappable(norm=normy, cmap=colormap.black_to_green)

    scalarMapx.set_array(dx)
    scalarMapy.set_array(dy)

    colorx = scalarMapx.to_rgba(dx)
    colory = scalarMapy.to_rgba(dy)

    color = np.array(colorx) + np.array(colory)
    color[:, :, 3] = 1.
    xmin, xmax, ymin, ymax = xmin + xbuffer, xmax - xbuffer, ymin + ybuffer, ymax - ybuffer
    return color, xmin, xmax, ymin, ymax


def plot_2d_color_scale(x1_min, x1_max, x2_min, x2_max, plot_point=None,
                        resolution=10, ax=None, x_buffer_prop=.1, y_buffer_prop=.1):
    """
    Return a generic plot of a 2D color scale

    Parameters
    ----------
    x1_min: numeric
        how high x1 should extend the color scale
    x1_max:  numeric
        how high 2 should extend the color scale
    x2_min:  numeric
        how low x2 should extend the color scale
    x2_max: numeric
        how high x2 should extend the color scale
    resolution: int
        how fine grain to make the color scale
    ax: matplotlib.axes._subplots.AxesSubplot
        matplotlib axis to plot on. if none will generate a new one

    Returns
    ----------
    matplotlib.axes._subplots.AxesSubplot
    """

    try:
        global pyplot
        global patches
        from matplotlib import pyplot, patches
    except ImportError:
        raise (MatplotlibUnavailableError("Matplotlib is required but unavailable on your system."))
    except RuntimeError:
        raise (MatplotlibDisplayError("Matplotlib unable to open display"))

    # xmin, xmax = min(0, np.percentile(x, 3)), max(np.percentile(x, 97), 0)

    if ax is None:
        f, ax = pyplot.subplots(1)

    x1_min, x1_max, x1_buffer = build_buffer(x1_min, x1_max, buffer_prop=x_buffer_prop)
    x2_min, x2_max, x2_buffer = build_buffer(x2_min, x2_max, buffer_prop=y_buffer_prop)
    ax.set_xlim(x1_min, x1_max)
    ax.set_ylim(x2_min, x2_max)
    x1 = np.linspace(x1_min, x1_max, resolution + 1)
    x2 = np.linspace(x2_min, x2_max, resolution + 1)
    x1_diff = x1[1] - x1[0]
    x2_diff = x2[1] - x2[0]
    x1, x2 = np.meshgrid(x1, x2)
    colors_for_scale, a, b, c, d = coordinate_gradients_to_1d_colorscale(x1,
                                                                         x2,
                                                                         x_buffer_prop=0,
                                                                         y_buffer_prop=0)

    if ax is None:
        f, ax = pyplot.subplots(1)
    for i in range(resolution):
        for j in range(resolution):
            xy = (x1[i, j], x2[i, j])
            color = colors_for_scale[i, j]
            rect = patches.Rectangle(
                xy, x1_diff, x2_diff,
                alpha=.8, facecolor=color, zorder=1
            )
            ax.add_patch(rect)
    if plot_point:
        ax.scatter([plot_point[0]], [plot_point[1]], s=75.,
                   color='yellow', alpha=1, zorder=2)
    ax.grid()
    return ax


def build_buffer(xmin, xmax, buffer_prop=.1):
    if xmin != xmax:
        buffer = (xmax - xmin) * buffer_prop / 2.
    # all values are the same
    else:
        buffer = 0.01
    xmin, xmax = xmin - buffer, xmax + buffer
    return xmin, xmax, buffer


def tick_formatter(powerlimits=None):
    try:
        from matplotlib.ticker import ScalarFormatter
    except ImportError:
        raise (MatplotlibUnavailableError("Matplotlib is required but unavailable on your system."))
    except RuntimeError:
        raise (MatplotlibDisplayError("Matplotlib unable to open display"))
    if powerlimits is None:
        powerlimits = (3, 3)
    formatter = ScalarFormatter()
    formatter.set_powerlimits(powerlimits)
    return formatter



class ControlledDict(dict):
    __readonly = False

    def block_setitem(self):
        """Allow or deny modifying dictionary"""
        self.__readonly = True

    def allow_setitem(self):
        """Allow or deny modifying dictionary"""
        self.__readonly = False

    def __setitem__(self, key, value):

        if self.__readonly:
            raise(TypeError, "__setitem__ is not supported")
        return dict.__setitem__(self, key, value)

    def __delitem__(self, key):

        if self.__readonly:
            raise(TypeError, "__delitem__ is not supported")
        return dict.__delitem__(self, key)



# Reference: http://python-packaging.readthedocs.io/en/latest/dependencies.html
from setuptools import setup, find_packages
import os, io, sys
import contextlib

@contextlib.contextmanager
def chdir(new_dir):
    old_dir = os.getcwd()
    try:
        os.chdir(new_dir)
        sys.path.insert(0, new_dir)
        yield
    finally:
        del sys.path[0]
        os.chdir(old_dir)

def setup_package():
    root = os.path.abspath(os.path.dirname(__file__))

    with chdir(root):
        with io.open(os.path.join(root, 'skater', 'about.py'), encoding='utf8') as f:
            about = {}
            exec(f.read(), about)

        with io.open(os.path.join(root, 'description.rst'), encoding='utf8') as f:
            readme = f.read()

    setup(
        name=about['__title__'],
        zip_safe=False,
        packages=find_packages(),
        description=about['__summary__'],
        long_description=readme,
        author=about['__author__'],
        author_email=about['__email__'],
        version=about['__version__'],
        url=about['__uri__'],
        license=about['__license__'],
        install_requires=[
            'scikit-learn>=0.18',
            'pandas>=0.19',
            'ds-lime>=0.1.1.21',
            'requests',
            'pathos==0.2.0',
            'dill>=0.2.6'],
        extras_require ={'all':'matplotlib'},
        )

if __name__ == '__main__':
    setup_package()



"""Base ModelInterpreter class. submodules like lime and partial dependence"
must have these methods"""


class ModelInterpreter(object):
    """
    Base Interpreter class. Common methods include loading a dataset and type setting.
    """

    def __init__(self, interpreter):
        self.interpreter = interpreter

    @property
    def data_set(self):
        """data_set routes to the Interpreter's dataset"""
        return self.interpreter.data_set

    @staticmethod
    def _types():
        return ['partial_dependence', 'lime']

    def load_data(self, training_data, index=None, feature_names=None):
        """.consider routes to Interpreter's .consider"""
        self.interpreter.consider(training_data, index=index, feature_names=feature_names)



from __future__ import division
import numpy as np
from sklearn.preprocessing import LabelBinarizer


def flatten(array):
    return [item for sublist in array for item in sublist]


class MultiColumnLabelBinarizer(LabelBinarizer):
    def __init__(self, neg_label=0, pos_label=1, sparse_output=False):
        self.neg_label = neg_label
        self.pos_label = pos_label
        self.sparse_output = sparse_output
        self.binarizers = []


    def fit(self, X):
        for x in X.T:
            binarizer = LabelBinarizer()
            binarizer.fit(x)
            self.binarizers.append(binarizer)


    def transform(self, X):
        results = []
        for i, x in enumerate(X.T):
            results.append(self.binarizers[i].transform(x))
        return np.concatenate(results, axis=1)


    def fit_transform(self, X):
        self.fit(X)
        return self.transform(X)


    def inverse_transform(self, X):
        results = []
        column_counter = 0

        for i, binarizer in enumerate(self.binarizers):
            n_cols = binarizer.classes_.shape[0]
            x_subset = X[:, column_counter:column_counter + n_cols]
            inv = binarizer.inverse_transform(x_subset)
            if len(inv.shape) == 1:
                inv = inv[:, np.newaxis]
            results.append(inv)
            column_counter += n_cols
        return np.concatenate(results, axis=1)


def add_column_numpy_array(array, new_col):
    placeholder = np.ones(array.shape[0])[:, np.newaxis]
    result = np.hstack((array, placeholder))

    if isinstance(new_col, np.ndarray):
        assert array.shape[0] == new_col.shape[0], "input array row counts \
                                                    must be the same. \
                                                    Expected: {0}\
                                                    Actual: {1}".format(array.shape[0],
                                                                        new_col.shape[0])
        assert len(new_col.shape) <= 2, "new column must be 1D or 2D"

        if len(new_col.shape) == 1:
            new_col = new_col[:, np.newaxis]
        return np.hstack((array, new_col))
    elif isinstance(new_col, list):
        assert len(new_col) == array.shape[0], "input array row counts \
                                                    must be the same. \
                                                    Expected: {0}\
                                                    Actual: {1}".format(len(array),
                                                                        len(new_col))
        new_col = np.array(new_col)
        assert len(new_col.shape) == 1, "list elements cannot be iterable"
        new_col = new_col[:, np.newaxis]
        return np.hstack((array, new_col))
    else:
        placeholder = np.ones(array.shape[0])[:, np.newaxis]
        result = np.hstack((array, placeholder))
        result[:, -1] = new_col
        return result


def allocate_samples_to_bins(n_samples, ideal_bin_count=100):
    """goal is as best as possible pick a number of bins
    and per bin samples to a achieve a given number
    of samples.

    Parameters
    ----------

    Returns
    ----------
    number of bins, list of samples per bin
    """

    if n_samples <= ideal_bin_count:
        n_bins = n_samples
        samples_per_bin = [1 for _ in range(n_bins)]
    else:
        n_bins = ideal_bin_count
        remainer = n_samples % ideal_bin_count

        samples_per_bin = np.array([(n_samples - remainer) / ideal_bin_count for _ in range(n_bins)])
        if remainer != 0:
            additional_samples_per_bin = distribute_samples(remainer, n_bins)
            samples_per_bin = samples_per_bin + additional_samples_per_bin
    return n_bins, samples_per_bin


def distribute_samples(n_samples, n_bins):
    assert n_samples < n_bins, "number of samples should be \
                                less than number of bins"
    space_size = n_bins / n_samples

    samples_per_bin = np.zeros(n_bins).tolist()

    index_counter = 0
    for sample in range(n_samples):
        index = int(index_counter)
        samples_per_bin[index] += 1
        index_counter += space_size

    return samples_per_bin


def divide_zerosafe(a, b):
    """ diving by zero returns 0 """
    with np.errstate(divide='ignore', invalid='ignore'):
        c = np.true_divide(a, b)
        c[~np.isfinite(c)] = 0  # -inf inf NaN

    return c



"""Funcs for logging"""
import logging


def build_logger(log_level, logger_name):
    logger = logging.Logger(logger_name)
    logger.setLevel(log_level)
    msg_formatter = logging.Formatter(
        '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
    )
    stream_handler = logging.StreamHandler()
    stream_handler.setFormatter(msg_formatter)
    stream_handler.setFormatter(msg_formatter)
    logger.addHandler(stream_handler)
    return logger



"""Model class."""

import abc
import numpy as np
from sklearn.preprocessing import LabelEncoder, OneHotEncoder
from sklearn.utils.multiclass import type_of_target
import pandas as pd

from ..util.static_types import StaticTypes
from ..util.logger import build_logger
from ..util import exceptions
from ..data import DataManager


class ModelType(object):
    """What is a model? A model needs to make predictions, so a means of
    passing data into the model, and receiving results.

    Goals:
        We want to abstract away how we access the model.
        We want to make inferences about the format of the output.
        We want to able to map model outputs to some smaller, universal set of output types.
        We want to infer whether the model is real valued, or classification (n classes?)
    """
    __metaclass__ = abc.ABCMeta


    def __init__(self, log_level=30, target_names=None, examples=None, feature_names=None, unique_values=None,
                 input_formatter=None, output_formatter=None):
        """
        Base model class for wrapping prediction functions. Common methods
        involve output type inference in requiring predict methods

        Parameters
        ----------
            log_level: int
                0, 10, 20, 30, 40, or 50 for verbosity of logs.
            target_names: arraytype
                The names of the target variable/classes. There should be as many
                 names as there are outputs per prediction (n=1 for regression,
                 n=2 for binary classification, etc). Defaults to Predicted Value for
                 regression and Class 1...n for classification.



        Attributes
        ----------
            model_type: string

        """
        self._log_level = log_level
        self.logger = build_logger(log_level, __name__)
        self.examples = None
        self.model_type = StaticTypes.unknown
        self.output_var_type = StaticTypes.unknown
        self.output_shape = StaticTypes.unknown
        self.n_classes = StaticTypes.unknown
        self.input_shape = StaticTypes.unknown
        self.probability = StaticTypes.unknown
        self.transformer = self.identity_function
        self.label_encoder = LabelEncoder()
        self.one_hot_encoder = OneHotEncoder()
        self.target_names = target_names
        self.feature_names = feature_names
        self.unique_values = unique_values
        self.input_formatter = input_formatter or self.identity_function
        self.output_formatter = output_formatter or self.identity_function
        self.has_metadata = False

        if examples is not None:
            self.input_type = type(examples)
            examples = DataManager(examples, feature_names=feature_names)
            self._build_model_metadata(examples)
        else:
            self.input_type = None
            self.logger.warn("No examples provided, cannot infer model type")


    def predict(self, *args, **kwargs):
        """
        The way in which the submodule predicts values given an input
        """
        return self.transformer(self.output_formatter(self._execute(self.input_formatter(*args, **kwargs))))


    @abc.abstractmethod
    def _execute(self, *args, **kwargs):
        """
        The way in which the submodule predicts values given an input
        """
        return


    @abc.abstractmethod
    def _predict(self, *args, **kwargs):
        """
        The way in which the submodule predicts values given an input
        """
        return

    @abc.abstractmethod
    def _get_static_predictor(self, *args, **kwargs):
        """Return a static prediction function to avoid shared state in multiprocessing"""
        return


    def __call__(self, *args, **kwargs):
        return self.predict(*args, **kwargs)


    def check_examples(self, examples):
        """
        Ties examples to self. equivalent to self.examples = np.array(examples).
        Parameters
        ----------
        examples: array type


        """
        if isinstance(examples, (pd.DataFrame, np.ndarray)):
            return examples
        else:
            return np.array(examples)


    def _build_model_metadata(self, dataset):
        """
        Determines the model_type, output_type. Side effects
        of this method are to mutate object's attributes (model_type,
        n_classes, etc).

        Parameters
        ----------
        examples: pandas.DataFrame or numpy.ndarray
            The examples that will be passed through the predict function.
            The outputs from these examples will be used to make inferences
            about the types of outputs the function generally makes.

        """
        self.logger.debug("Beginning output checks")

        if self.input_type in (None, pd.DataFrame):
            outputs = self.predict(dataset.data)
        elif self.input_type == np.ndarray:
            outputs = self.predict(dataset.data)
        else:
            raise ValueError("Unrecognized input type: {}".format(self.input_type))

        self.input_shape = dataset.data.shape
        self.output_shape = outputs.shape

        ndim = len(outputs.shape)
        if ndim > 2:
            raise(ValueError("Unsupported model type, output dim = {}".format(ndim)))

        try:
            # continuous, binary, continuous multioutput, multiclass, multilabel-indicator
            self.output_type = type_of_target(outputs)
        except:
            self.output_type = False

        if self.output_type == 'continuous':
            self.model_type = StaticTypes.model_types.regressor
            self.n_classes = 1
            self.probability = StaticTypes.not_applicable

        elif self.output_type == 'multiclass':
            self.model_type = StaticTypes.model_types.classifier
            self.probability = False
            self.n_classes = len(np.unique(outputs))

        elif self.output_type == 'continuous-multioutput':
            self.model_type = StaticTypes.model_types.classifier
            self.probability = True
            self.n_classes = outputs.shape[1]

        elif self.output_type == 'binary':
            self.model_type = StaticTypes.model_types.classifier
            self.probability = False
            self.n_classes = 2

        elif self.output_type == 'multilabel-indicator':
            self.model_type = StaticTypes.model_types.classifier
            self.probability = False
            self.n_classes = outputs.shape[1]

        else:
            err_msg = "Could not infer model type"
            self.logger.debug("Inputs: {}".format(dataset.data))
            self.logger.debug("Outputs: {}".format(outputs))
            self.logger.debug("sklearn response: {}".format(self.output_type))
            exceptions.ModelError(err_msg)

        if self.target_names is None:
            self.target_names = range(self.n_classes)

        if self.unique_values is None and self.model_type == 'classifier' and self.probability is False:
            raise (exceptions.ModelError('If using classifier without probability scores, unique_values cannot '
                                         'be None'))

        self.transformer = self.transformer_func_factory(outputs)

        reports = self.model_report(dataset.data)
        for report in reports:
            self.logger.debug(report)

        if self.target_names is None:
            self.target_names = range(self.n_classes)

        self.has_metadata = True


    def predict_function_transformer(self, output):
        """
        Call this method when model returns a 1D array of
        predicted classes. The output is one hot encoded version.

        Parameters
        ----------
        output: array type
            The output of the pre-formatted predict function

        Returns
        ----------
        output: numpy.ndarray
            The one hot encoded outputs of predict_fn
        """

        _labels = self.label_encoder.transform(output)[:, np.newaxis]
        # target_names = label_encoder.classes_.tolist()

        self.logger.debug("Using transforming function. Found {} classes".format(len(self.label_encoder.classes_)))
        self.logger.debug("Label shape: {}".format(len(_labels.shape)))
        output = self.one_hot_encoder.transform(_labels).todense()
        output = np.squeeze(np.asarray(output))
        return DataManager(output, feature_names=self.label_encoder.classes_)[self.unique_values]


    def transformer_func_factory(self, outputs):
        """
        In the event that the predict func returns 1D array of predictions,
        then this returns a formatter to convert outputs to a 2D one hot encoded
        array.

        For instance, if:
            predict_fn(data) -> ['apple','banana']
        then
            transformer = Model.transformer_func_factory()
            transformer(predict_fn(data)) -> [[1, 0], [0, 1]]

        Returns
        ----------
        (callable):
            formatter function to wrap around predict_fn
        """

        # Note this expression below assumptions (not probability) evaluates to false if
        # and only if the model does not return probabilities. If unknown, should be true
        if self.model_type == StaticTypes.model_types.classifier and not self.probability:
            # fit label encoder
            artificial_samples = np.array(self.unique_values)
            self.logger.debug("Label encoder fit on examples of shape: {}".format(outputs.shape))
            self.label_encoder.fit(artificial_samples)
            labels = self.label_encoder.transform(artificial_samples)[:, np.newaxis]
            self.logger.debug("Onehot encoder fit on examples of shape: {}".format(labels.shape))
            self.one_hot_encoder.fit(labels)
            return self.predict_function_transformer
        else:
            return lambda x: x


    def model_report(self, examples):
        """
        Just returns a list of model attributes as a list

        Parameters
        ----------
        examples: array type:
            Examples to use for which we report behavior of predict_fn.


        Returns
        ----------
        reports: list of strings
            metadata about function.

        """
        examples = DataManager(examples, feature_names=self.feature_names)
        reports = []
        if isinstance(self.examples, np.ndarray):
            raw_predictions = self.predict(examples)
            reports.append("Example: {} \n".format(examples[0]))
            reports.append("Outputs: {} \n".format(raw_predictions[0]))
        reports.append("Model type: {} \n".format(self.model_type))
        reports.append("Output Var Type: {} \n".format(self.output_var_type))
        reports.append("Output Shape: {} \n".format(self.output_shape))
        reports.append("N Classes: {} \n".format(self.n_classes))
        reports.append("Input Shape: {} \n".format(self.input_shape))
        reports.append("Probability: {} \n".format(self.probability))
        return reports

    def predict_subset_classes(self, data, subset_of_classes):
        """Filters predictions to a subset of classes."""
        if subset_of_classes is None:
            return self.predict(data)
        else:
            return DataManager(self.predict(data), feature_names=self.target_names)[subset_of_classes]


    @staticmethod
    def identity_function(x):
        return x



"""Model subclass for in memory predict functions"""
from functools import partial

from .base import ModelType
from ..util import exceptions


class InMemoryModel(ModelType):
    """
    This model can be called directly from memory
    """

    def __init__(self, prediction_fn, input_formatter=None,
                 output_formatter=None, target_names=None,
                 unique_values=None, examples=None, log_level=30):
        """This model can be called directly from memory

        Parameters
        ----------
        prediction_fn: callable
            function that returns predictions

        input_formatter: callable
            This function will run on input data before passing
            to the prediction_fn. This usually should take your data type
            and convert them to numpy arrays or dataframes.

        output_formatter: callable
            This function will run on input data before passing
            to the prediction_fn. This usually should take your data type
            and convert them to numpy arrays or dataframes.

        target_names: array type
            optional names of classes that describe model outputs.

        unique_values: array type
            The set of possible output values. Only use on classifier models that
            return "best guess" predictions, not probability scores, e.g.

            model.predict(fruit1) -> 'apple'
            model.predict(fruit2) -> 'banana'

            ['apple','banana'] are the unique_values of the classifier

        examples: numpy.array or pandas.dataframe
            optional examples to use to make inferences about the function.

        log_level: int
            config setting to see model logs. 10 is a good value for seeing debug messages.
            30 is warnings only.


        """

        if not hasattr(prediction_fn, "__call__"):
            raise(exceptions.ModelError("Predict function must be callable"))

        self.prediction_fn = prediction_fn
        super(InMemoryModel, self).__init__(log_level=log_level,
                                            target_names=target_names,
                                            examples=examples,
                                            unique_values=unique_values,
                                            input_formatter=input_formatter,
                                            output_formatter=output_formatter)


    def _execute(self, *args, **kwargs):
        """
        Just use the function itself for predictions
        """
        return self.prediction_fn(*args, **kwargs)


    @staticmethod
    def _predict(data, predict_fn, input_formatter, output_formatter, transformer=None):
        """Static prediction function for multiprocessing usecases

        Parameters
        ----------
        data: arraytype

        formatter: callable
            function responsible for formatting model outputs as necessary. For instance,
            one hot encoding multiclass outputs.

        predict_fn: callable

        Returns
        -----------
        predictions: arraytype
        """
        results = output_formatter(predict_fn(input_formatter(data)))
        if transformer:
            return transformer(results)
        else:
            return results

    def _get_static_predictor(self):
        transformer = self.transformer
        input_formatter = self.input_formatter
        output_formatter = self.output_formatter
        prediction_fn = self.prediction_fn
        predict_fn = partial(self._predict,
                             transformer=transformer,
                             predict_fn=prediction_fn,
                             input_formatter=input_formatter,
                             output_formatter=output_formatter,
                             )
        return predict_fn



import unittest

import numpy as np
from scipy.stats import norm
from scipy.special import expit
from sklearn.linear_model import LinearRegression, LogisticRegression
from sklearn.ensemble import GradientBoostingClassifier
from sklearn import datasets
from functools import partial

from skater.core.explanations import Interpretation
from skater.util import exceptions
from arg_parser import create_parser
from skater.model import InMemoryModel, DeployedModel
from skater.util.dataops import MultiColumnLabelBinarizer
from skater.core.global_interpretation.partial_dependence import PartialDependence


class TestPartialDependence(unittest.TestCase):

    def setUp(self):
        args = create_parser().parse_args()
        debug = args.debug
        self.seed = args.seed
        self.n = args.n
        self.dim = args.dim
        self.features = [str(i) for i in range(self.dim)]
        self.X = norm.rvs(0, 1, size=(self.n, self.dim), random_state=self.seed)
        self.B = np.array([-10.1, 2.2, 6.1])
        self.y = np.dot(self.X, self.B)
        self.y_as_int = np.round(expit(self.y))
        self.y_as_string = np.array([str(i) for i in self.y_as_int])
        # example dataset for y = B.X
        # X = array([[ 1.62434536, -0.61175641, -0.52817175], ... [-0.15065961, -1.40002289, -1.30106608]])  (1000 * 3)
        # B = array([-10.1,   2.2,   6.1])
        # y = array([ -2.09736000e+01,  -1.29850618e+00,  -1.73511155e+01, ...]) (1000 * 1)
        # features = ['0', '1', '2']
        ##
        # Other output types:
        # y_as_int = array[ 0.,  0.,  0.,  0.,  1.,  1.,  0.,  0.,  0.,  1.,  1.,  1.,  1., ...]
        # y_as_string = array['0.0', '0.0', '0.0', '0.0', '1.0', '1.0', '0.0', '0.0', '0.0', ... ]


        # Another set of input
        # sample data
        self.sample_x = np.array([[-2, -1], [-1, -1], [-1, -2], [1, 1], [1, 2], [2, 1]])
        self.sample_y = np.array([-1, -1, -1, 1, 1, 1])
        self.sample_feature_name = [str(i) for i in range(self.sample_x.shape[1])]

        if debug:
            self.interpreter = Interpretation(log_level='DEBUG')
        else:
            self.interpreter = Interpretation()  # default level is 'WARNING'
        self.interpreter.load_data(self.X, feature_names=self.features)

        self.regressor = LinearRegression()
        self.regressor.fit(self.X, self.y)
        self.regressor_predict_fn = InMemoryModel(self.regressor.predict, examples=self.X)

        self.classifier = LogisticRegression()
        self.classifier.fit(self.X, self.y_as_int)
        self.classifier_predict_fn = InMemoryModel(self.classifier.predict, examples=self.X, unique_values=self.classifier.classes_)
        self.classifier_predict_proba_fn = InMemoryModel(self.classifier.predict_proba, examples=self.X)

        self.string_classifier = LogisticRegression()
        self.string_classifier.fit(self.X, self.y_as_string)
        self.string_classifier_predict_fn = InMemoryModel(self.string_classifier.predict_proba, examples=self.X)


        # Yet another set of input!!
        self.sample_x_categorical = np.array([['B', -1], ['A', -1], ['A', -2], ['C', 1], ['C', 2], ['A', 1]])
        self.sample_y_categorical = np.array(['A', 'A', 'A', 'B', 'B', 'B'])
        self.categorical_feature_names = ['Letters', 'Numbers']
        self.categorical_transformer = MultiColumnLabelBinarizer()
        self.categorical_transformer.fit(self.sample_x_categorical)
        self.sample_x_categorical_transormed = self.categorical_transformer.transform(self.sample_x_categorical)
        self.categorical_classifier = LogisticRegression()
        self.categorical_classifier.fit(self.sample_x_categorical_transormed, self.sample_y_categorical)
        self.categorical_predict_fn = lambda x: self.categorical_classifier.predict_proba(self.categorical_transformer.transform(x))
        self.categorical_model = InMemoryModel(self.categorical_predict_fn, examples=self.sample_x_categorical)


    def test_pdp_with_default_sampling(self):
        pdp_df = self.interpreter.partial_dependence.partial_dependence([self.features[0]],
                                                                        self.regressor_predict_fn,
                                                                        sample=True)
        self.assertEquals(pdp_df.shape, (30, 3))  # default grid resolution is 30

    def test_pd_with_categorical_features(self):
        interpreter = Interpretation(self.sample_x_categorical, feature_names=self.categorical_feature_names)
        try:
            interpreter.partial_dependence.partial_dependence([self.categorical_feature_names[0]], self.categorical_model)
        except:
            self.fail("PD computation function failed with categorical features")
        try:
            interpreter.partial_dependence.plot_partial_dependence([self.categorical_feature_names], self.categorical_model)
        except:
            self.fail("PDP plotting function failed with categorical features")



    def test_partial_dependence_binary_classification(self):
        # In the default implementation of pdp on sklearn, there is an approx. done
        # if the number of unique values for a feature space < grid_resolution specified.
        # For now, we have decided to not have that approximation. In V2, we will be benchmarking for
        # performance as well. Around that time we will revisit the same.
        # Reference: https://github.com/scikit-learn/scikit-learn/blob/4d9a12d175a38f2bcb720389ad2213f71a3d7697/sklearn/ensemble/tests/test_partial_dependence.py
        # TODO: check on the feature space approximation (V2)
        # Test partial dependence for classifier
        clf = GradientBoostingClassifier(n_estimators=10, random_state=1)
        clf.fit(self.sample_x, self.sample_y)
        classifier_predict_fn = InMemoryModel(clf.predict_proba, examples=self.sample_x)
        interpreter = Interpretation()
        interpreter.load_data(np.array(self.sample_x), self.sample_feature_name)
        pdp_df = interpreter.partial_dependence.partial_dependence(['0'],
                                                                   classifier_predict_fn,
                                                                   grid_resolution=5,
                                                                   sample=True)

        self.assertEquals(pdp_df.shape[0], len(np.unique(interpreter.data_set['0'])))

        # now with our own grid
        ud_grid = np.unique(self.sample_x[:, 0])
        # input: array([-2, -1,  1,  2])
        # the returned grid should have only 4 values as specified by the user
        pdp_df = interpreter.partial_dependence.partial_dependence(['0'], classifier_predict_fn,
                                                                   grid=ud_grid, sample=True)
        self.assertEquals(pdp_df.shape[0], 4)


    def test_partial_dependence_multiclass(self):
        # Iris data classes: ['setosa', 'versicolor', 'virginica']
        iris = datasets.load_iris()
        # 1. Using GB Classifier
        clf = GradientBoostingClassifier(n_estimators=10, random_state=1)
        clf.fit(iris.data, iris.target)
        classifier_predict_fn = InMemoryModel(clf.predict_proba, examples=iris.data)
        interpreter = Interpretation()
        interpreter.load_data(iris.data, iris.feature_names)
        pdp_df = interpreter.partial_dependence.partial_dependence([iris.feature_names[0]], classifier_predict_fn,
                                                                   grid_resolution=25, sample=True)

        expected_feature_name = PartialDependence.feature_column_name_formatter('sepal length (cm)')

        self.assertIn(expected_feature_name,
                      pdp_df.columns.values,
                      "{0} not in columns {1}".format(expected_feature_name,
                                                      pdp_df.columns.values))
        # 2. Using SVC
        from sklearn import svm
        # With SVC, predict_proba is supported only if probability flag is enabled, by default it is false
        clf = svm.SVC(probability=True)
        clf.fit(iris.data, iris.target)
        classifier_predict_fn = InMemoryModel(clf.predict_proba, examples=iris.data)
        interpreter = Interpretation()
        interpreter.load_data(iris.data, iris.feature_names)
        pdp_df = interpreter.partial_dependence.partial_dependence([iris.feature_names[0]], classifier_predict_fn,
                                                                   grid_resolution=25, sample=True)
        self.assertIn(expected_feature_name,
                      pdp_df.columns.values,
                      "{} not in columns {}".format(*[expected_feature_name,
                                                      pdp_df.columns.values]))




    def test_pdp_regression_coefs_closeness(self, epsilon=1):
        pdp_df = self.interpreter.partial_dependence.partial_dependence([self.features[0]],
                                                                        self.regressor_predict_fn)
        val_col = PartialDependence.feature_column_name_formatter(self.features[0])

        y = np.array(pdp_df[self.regressor_predict_fn.target_names[0]])
        x = np.array(pdp_df[val_col])[:, np.newaxis]
        regressor = LinearRegression()
        regressor.fit(x, y)
        self.interpreter.logger.debug("Regressor coefs: {}".format(regressor.coef_))
        self.interpreter.logger.debug("Regressor coef shape: {}".format(regressor.coef_.shape))
        coef = regressor.coef_[0]
        self.assertTrue(abs(coef - self.B[0]) < epsilon, True)


    def test_pdp_inputs(self):
        clf = GradientBoostingClassifier(n_estimators=10, random_state=1)
        clf.fit(self.sample_x, self.sample_y)
        interpreter = Interpretation()
        self.assertRaisesRegexp(Exception, "Invalid Data", interpreter.load_data, None, self.sample_feature_name)


    def test_2D_pdp(self):
        try:
            self.interpreter.partial_dependence.partial_dependence(self.features[:2],
                                                                   self.regressor_predict_fn,
                                                                   grid_resolution=10,
                                                                   sample=True)
        except:
            self.fail("2D regressor pd failed")


    def test_plot_1D_pdp(self):
        try:
            self.interpreter.partial_dependence.plot_partial_dependence([self.features[0]],
                                                                        self.regressor_predict_fn,
                                                                        grid_resolution=10)
        except:
            self.fail("1D regressor plot failed")


    def test_plot_1D_pdp_with_sampling(self):
        try:
            self.interpreter.partial_dependence.plot_partial_dependence(
                [self.features[0]],
                self.regressor_predict_fn,
                grid_resolution=10,
                sample=True)
        except:
            self.fail("1D classifier plot with sampling failed")


    def test_plot_2D_pdp(self):
        try:
            self.interpreter.partial_dependence.plot_partial_dependence(self.features[:2],
                                                                        self.regressor_predict_fn,
                                                                        grid_resolution=10,
                                                                        sample=False)
        except:
            self.fail("2D partial dep plot failed")

    def test_plot_2D_pdp_with_sampling(self):
        try:
            self.interpreter.partial_dependence.plot_partial_dependence(self.features[:2],
                                                                        self.regressor_predict_fn,
                                                                        grid_resolution=10,
                                                                        sample=True)
        except:
            self.fail("2D regressor with sampling failed")


    def test_fail_when_grid_range_is_outside_0_and_1(self):
        pdp_func = partial(self.interpreter.partial_dependence.partial_dependence,
                           *[[self.features[0]], self.regressor_predict_fn],
                           **{'grid_range': (.01, 1.01)})
        self.assertRaises(exceptions.MalformedGridRangeError, pdp_func)


    def test_pdp_1d_classifier_no_proba(self):
        self.interpreter.partial_dependence.partial_dependence(self.features[:1],
                                                               self.classifier_predict_fn,
                                                               grid_resolution=10)
        try:
            self.interpreter.partial_dependence.partial_dependence(self.features[:1],
                                                                   self.classifier_predict_fn,
                                                                   grid_resolution=10)
        except:
            self.fail("1D pdp without proba failed")


    def test_pdp_2d_classifier_no_proba(self):
        try:
            self.interpreter.partial_dependence.partial_dependence(self.features[:2],
                                                                   self.classifier_predict_fn,
                                                                   grid_resolution=10)
        except:
            self.fail("2D pdp without proba failed")


    def test_pdp_1d_classifier_with_proba(self):
        try:
            self.interpreter.partial_dependence.partial_dependence(self.features[:1],
                                                                   self.classifier_predict_proba_fn,
                                                                   grid_resolution=10)
        except:
            self.fail("1D classifier with probability scores failed")


    def test_pdp_2d_classifier_with_proba(self):
        try:
            self.interpreter.partial_dependence.partial_dependence(self.features[:2],
                                                                   self.classifier_predict_proba_fn,
                                                                   grid_resolution=10)
        except:
            self.fail("2D classifier with probability scores failed")


    def test_pdp_1d_string_classifier_no_proba(self):
        def fail_func():
            self.interpreter.partial_dependence.partial_dependence(self.features[:1],
                                                                   InMemoryModel(self.string_classifier.predict,
                                                                                 examples=self.X),
                                                                   grid_resolution=10)
        self.assertRaises(exceptions.ModelError, fail_func)


    def test_pdp_1d_string_classifier_with_proba(self):
        try:
            self.interpreter.partial_dependence.partial_dependence(self.features[:1],
                                                                   self.string_classifier_predict_fn,
                                                                   grid_resolution=10)
        except:
            self.fail('1D string classifier pd failed')



if __name__ == '__main__':
    suite = unittest.TestLoader().loadTestsFromTestCase(TestPartialDependence)
    unittest.TextTestRunner(verbosity=2).run(suite)



"""The modules responsible for wrapping models through a common interface. To enable support
fot classification models with heterogenous output types, and for deployed models that are
accessible through api calls, we need to contruct various patterns of prediction pipelines
and of converting data throughout. Once a prediction pipeline is defined, we use the base Model class
to define how to type inference.
"""

from .local_model import InMemoryModel
from .deployed_model import DeployedModel

__all__ = ['InMemoryModel', 'DeployedModel']
